# Copyright (C) 2023 Maxwell G <gotmax@e.email>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import abc
import argparse
import logging
import warnings
from collections.abc import Callable, ItemsView, Iterable, Iterator, Mapping
from contextlib import suppress
from typing import TYPE_CHECKING, Any, NoReturn

from fedrq._utils import get_source_name

if TYPE_CHECKING:
    from fedrq.backends.base import PackageCompat

LOG = logging.getLogger(__name__)


class FormatterError(Exception):
    pass


_ATTRS: tuple[str, ...] = (
    "name",
    "arch",
    "a",
    "epoch",
    "e",
    "version",
    "v",
    "release",
    "r",
    "from_repo",
    "evr",
    "debug_name",
    "source_name",
    "source_debug_name",
    "installtime",
    "buildtime",
    "size",
    "downloadsize",
    "installsize",
    "provides",
    "requires",
    "recommends",
    "suggests",
    "supplements",
    "enhances",
    "obsoletes",
    "conflicts",
    "sourcerpm",
    "description",
    "summary",
    "license",
    "url",
    "reason",
    "files",
    "reponame",
    "repoid",
    "vendor",
    "packager",
    "location",
)


def _stringify(value: Any, *, multiline_allowed: bool = True) -> str:
    if value is None or value == "":
        return "(none)"
    if isinstance(value, str) and "\n" in value:
        if not multiline_allowed:
            raise FormatterError("Multiline values are not allowed")
        return value + "\n---\n"
    return str(value)


class Formatter(abc.ABC):
    ATTRS = _ATTRS
    MULTILINE = False

    """
    Convert PackageCompat objects into a string representation
    for use with the fedrq CLI.
    """

    def __init__(
        self,
        name: str,
        seperator: str,
        args: str,
        container: Formatters | None = None,
    ) -> None:
        self.name = name
        self.seperator = seperator
        self.args = args
        self.container: Formatters = container or Formatters({})
        self.validate()

    @abc.abstractmethod
    def format_line(self, package: PackageCompat) -> str:
        """
        Format a single Package object
        """
        pass

    def format(self, packages: Iterable[PackageCompat]) -> Iterable[str]:
        """
        Convert an Iterable of PackageCompat objects
        (or a PackageQueryCompat object) to an Iterable of str.
        """
        yield from map(self.format_line, sorted(packages))

    def err(self, msg: str) -> NoReturn:
        raise FormatterError(f"{self.name!r} FormatterError: {msg}")

    def validate(self) -> None:
        if self.seperator:
            self.err("no arguments are accepted")

    @classmethod
    def frompat(cls, fmt: str, name: str) -> type[Formatter]:
        """
        Convert Python format string (e.g. `{0.name}.{0.arch}`)
        in a Formatter class.
        """

        def format_line(self, package):  # noqa: ARG001
            return fmt.format(package)

        dct = dict(
            __doc__=f"{name} -- {fmt}",
            __module__=__name__,
            format_line=format_line,
        )
        typ = type(f"{name.upper()}Formatter", (cls,), dct)
        return typ

    @classmethod
    def fromfunc(
        cls, func: Callable[[PackageCompat], str], name: str
    ) -> type[Formatter]:
        """
        Convert a function (or other callable) into a Formatter class.
        The function is used as the format_line() method.
        """
        dct = dict(
            format_line=func,
            __module__=__name__,
            __doc__=func.__doc__ if func.__doc__ else name,
        )
        typ = type(f"{name.upper()}Formatter", (cls,), dct)
        return typ

    def __str__(self) -> str:
        return f"{self.name}{self.seperator}{self.args}"


class Formatters(Mapping[str, type[Formatter]]):
    __slots__ = ("__data", "fallback")

    """
    Immutable mapping like class of Formatter classes.
    Converts strings and simple functions to Formatter objects.
    Allows merging and adding other Formatters.
    """

    def __init__(
        self,
        formatters: Mapping[str, Callable | type[Formatter] | str],
        fallback: type[Formatter] | None = None,
    ) -> None:
        self.__data = self._formattersv(dict(formatters))
        self.fallback: type[Formatter] | None = fallback

    def get_formatter(self, key: str) -> Formatter:
        name, seperator, args = key.partition(":")
        with suppress(KeyError):
            return self[name](name, seperator, args, self)
        if self.fallback:
            with suppress(FormatterError):
                return self.fallback(name, seperator, args, self)
        raise FormatterError(f"{key!r} is not a valid formatter")

    def __getitem__(self, key: str) -> type[Formatter]:
        return self.__data[key]

    def __len__(self) -> int:
        return len(self.__data)

    def __iter__(self) -> Iterator[str]:
        return iter(self.__data)

    def __or__(
        self, other: Mapping[str, Callable | type[Formatter] | str]
    ) -> Formatters:
        fallback: type[Formatter] | None = self.fallback
        if hasattr(other, "fallback"):
            fallback = other.fallback
        return type(self)({**self, **other}, fallback)

    def __contains__(self, name: object) -> bool:
        if not isinstance(name, str):
            raise TypeError
        try:
            self.get_formatter(name)
        except FormatterError:
            return False
        else:
            return True

    def singleline(self) -> Formatters:
        return Formatters(
            {
                name: formatter
                for name, formatter in self.items()
                if not formatter.MULTILINE
            }
        )

    def new(self, other: Mapping[str, str | Callable | type[Formatter]]) -> Formatters:
        return self | other

    def items(self) -> ItemsView[str, type[Formatter]]:
        return ItemsView(self)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.__data!r})"

    def _formattersv(self, formatters) -> dict[str, type[Formatter]]:
        if not formatters:
            return {}
        for name, formatter in formatters.items():
            if isinstance(formatter, type) and issubclass(formatter, Formatter):
                pass
            elif isinstance(formatter, str):
                formatters[name] = Formatter.frompat(formatter, name)
            elif callable(formatter):
                formatters[name] = Formatter.fromfunc(
                    formatter,  # type: ignore[arg-type]
                    name,
                )
            else:
                raise TypeError
        return formatters

    def _argcompleter(
        self,
        *,
        prefix: str,  # noqa: ARG002
        action: argparse.Action,  # noqa: ARG002
        parser: argparse.ArgumentParser,  # noqa: ARG002
        parsed_args: argparse.Namespace,  # noqa: ARG002
    ) -> list[str]:
        opts: list[str] = [*_ATTRS]
        for name, formatter in self.items():
            if issubclass(formatter, SpecialFormatter):
                opts.append(f"{name}:")
            else:
                opts.append(name)
        return opts


class SourceFromatter(Formatter):
    def format(self, packages: Iterable[PackageCompat]) -> Iterable[str]:
        return sorted(set(map(self.format_line, packages)))

    def format_line(self, package: PackageCompat) -> str:
        return get_source_name(package)


class SpecialFormatter(Formatter):
    """
    Formatter that accepts arguments
    """

    ATTRS: tuple[str, ...] = Formatter.ATTRS

    def _get_attrs(self, args: str) -> Iterator[str | Formatter]:
        for attr in args.split(","):
            if attr in self.ATTRS:
                yield attr
            elif attr in self.container.singleline():
                yield self.container.get_formatter(attr)
            else:
                self.err(f"invalid argument {attr!r}")

    def validate(self) -> None:
        if not self.args.strip() or self.args.strip() == ",":
            self.err("received less than 1 argument")


class AttrFormatter(SpecialFormatter):
    MULTILINE = True

    def validate(self) -> None:
        super().validate()
        self._validate()

    def _validate(self) -> None:
        if self.args not in self.ATTRS:
            raise FormatterError(f"'{self.args}' is not a valid attribute")
        self.attr = self.args

    def format_line(self, package: PackageCompat) -> str:
        # Return one string if there's multiple lines
        return "\n".join(self.format([package]))

    def format(self, packages: Iterable[PackageCompat]) -> Iterable[str]:
        for p in sorted(packages):
            result = getattr(p, self.attr)
            if isinstance(result, Iterable) and not isinstance(result, str):
                yield from map(_stringify, result)
            else:
                yield _stringify(result)


class AttrFallbackFormatter(AttrFormatter):
    def validate(self) -> None:
        if self.seperator:
            self.err("no arguments are accepted")
        self.args = self.name
        self._validate()


class JsonFormatter(SpecialFormatter):
    MULTILINE = True

    def validate(self) -> None:
        super().validate()
        self.attrs: list[Formatter | str] = list(self._get_attrs(self.args))

    def _format(self, package: PackageCompat) -> Iterable[tuple[str, Any]]:
        for attr in self.attrs:
            if isinstance(attr, str):
                result = getattr(package, attr)
                if isinstance(result, Iterable) and not isinstance(result, str):
                    result = [str(i) for i in result]
            else:
                result = attr.format_line(package)
            yield str(attr), result

    def format(self, packages: Iterable[PackageCompat]):
        import json

        data = [dict(self._format(package)) for package in packages]
        yield json.dumps(data, indent=2)

    def format_line(self, package: PackageCompat):
        raise NotImplementedError


class SingleLineFormatter(SpecialFormatter):
    DEFAULT_DIVIDER = " : "
    ATTRS: tuple[str, ...] = tuple(
        set(SpecialFormatter.ATTRS)
        - {
            "description",
            "provides",
            "requires",
            "recommends",
            "suggests",
            "supplements",
            "enhances",
            "obsoletes",
            "conflicts",
            "files",
        }
    )

    def validate(self) -> None:
        args, seperator, args2 = self.args.rpartition(":")
        LOG.debug(
            "parsing %s args: %r, %r, %r",
            "SingleLineFormatter.validate",
            args,
            seperator,
            args2,
        )
        # line:name:| -> params=name, divider = |
        if args and seperator and args2:
            LOG.debug("case 1")
            self.params = args
            self.divider = args2
        # line:name -> params=name, divider = DEFAULT_DIVIDER
        # line: -> params='', divider=DEFAULT_DIVIDER -> error in _get_attr()
        # line:: -> params='', divider=DEFAULT_DIVIDER -> error in _get_attr()
        elif not args:
            LOG.debug("case 2")
            self.params = args2
            self.divider = self.DEFAULT_DIVIDER
        # line:name: (trailing :) -> params=name, divider=DEFAULT_DIVIDER
        # line:otherformatter:arg: -> params=otherformatter:args,
        #                             divider=DEFAULT_DIVIDER
        # line::: -> error
        elif args and seperator and not args2:
            LOG.debug("case 3")
            self.params = args
            self.divider = self.DEFAULT_DIVIDER
        else:
            raise RuntimeError
        self.attrs = list(self._get_attrs(self.params))

    def format_line(self, package: PackageCompat) -> str:
        out = ""
        for index, attr in enumerate(self.attrs):
            if isinstance(attr, str):
                out += _stringify(getattr(package, attr), multiline_allowed=False)
            else:
                out += attr.format_line(package)
            if index != len(self.attrs) - 1:
                out += self.divider
        return out


class _DefaultFormatters(Formatters):
    def __call__(self) -> Any:
        warnings.warn(
            "DEPRECATED since 0.4.0: DefaultFormatters no longer needs to be called."
            " It's already initialized."
            " Just call DefaultFormatters.get_formatter() directly."
        )
        return self


DefaultFormatters = _DefaultFormatters(
    dict(
        plain="{0}",
        plainwithrepo="{0} {0.reponame}",
        nevrr="{0} {0.reponame}",
        na="{0.name}.{0.arch}",
        nev="{0.name}-{0.epoch}:{0.version}",
        nevr="{0.name}-{0.evr}",
        nevra="{0}",
        full_nevra="{0.name}-{0.evr}.{0.arch}",
        nv="{0.name}-{0.version}",
        source=SourceFromatter,
        src=SourceFromatter,
        attr=AttrFormatter,
        json=JsonFormatter,
        line=SingleLineFormatter,
    ),
    AttrFallbackFormatter,
)



# SPDX-FileCopyrightText: 2023 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later

"""
This module contains a fedrq backend (i.e. an implementation of the
[`fedrq.backends.base.BackendMod`][fedrq.backends.base.BackendMod] interface)
that uses the libdnf5 Python bindings.
"""

from __future__ import annotations

import functools
import inspect
import logging
import sys
import typing as t
from collections.abc import Collection, Iterable

from fedrq._utils import filter_latest
from fedrq.backends import MissingBackendError
from fedrq.backends.base import BackendMod, BaseMakerBase, RepoqueryBase
from fedrq.backends.libdnf5 import BACKEND  # noqa: F401

if t.TYPE_CHECKING:
    from _typeshed import StrPath
    from typing_extensions import TypeAlias, Unpack


try:
    import libdnf5
    import rpm
except ImportError as exc:
    raise MissingBackendError(str(exc)) from None

LOG = logging.getLogger(__name__)
Priority_RUNTIME = libdnf5.conf.Option.Priority_RUNTIME
StrIter = t.Union[list[str], tuple[str], str]
IntIter = t.Union[list[int], tuple[int], int]
CONVERT_TO_LIST = (str, int)


class _QueryFilterKwargs(t.TypedDict, total=False):
    name: t.Union[StrIter, libdnf5.rpm.PackageSet]
    name__eq: t.Union[StrIter, libdnf5.rpm.PackageSet]
    name__neq: t.Union[StrIter, libdnf5.rpm.PackageSet]
    name__glob: StrIter
    name__contains: StrIter

    epoch: t.Union[StrIter, IntIter]
    epoch__eq: t.Union[StrIter, IntIter]
    epoch__neq: t.Union[StrIter, IntIter]
    epoch__glob: StrIter
    epoch__gt: IntIter
    epoch__lt: IntIter
    epoch__lte: IntIter

    epoch__gte: IntIter
    version: StrIter
    version__eq: StrIter
    version__neq: StrIter
    version__glob: StrIter
    version__gt: StrIter
    version__lt: StrIter
    version__gte: StrIter
    version__lte: StrIter

    release: StrIter
    release__eq: StrIter
    release__neq: StrIter
    release__glob: StrIter
    release__gt: StrIter
    release__lt: StrIter
    release__gte: StrIter
    release__lte: StrIter

    arch: StrIter
    arch__eq: StrIter
    arch__neq: StrIter
    arch__glob: StrIter

    sourcerpm: StrIter
    sourcerpm__eq: StrIter
    sourcerpm__neq: StrIter
    sourcerpm__glob: StrIter

    url: StrIter
    url__eq: StrIter
    url__neq: StrIter
    url__glob: StrIter
    url__contains: StrIter

    summary: StrIter
    summary__eq: StrIter
    summary__neq: StrIter
    summary__glob: StrIter
    summary__contains: StrIter

    description: StrIter
    description__eq: StrIter
    description__neq: StrIter
    description__glob: StrIter
    description__contains: StrIter

    provides: t.Union[StrIter, libdnf5.rpm.ReldepList]
    provides__eq: t.Union[StrIter, libdnf5.rpm.ReldepList]
    provides__neq: t.Union[StrIter, libdnf5.rpm.ReldepList]
    provides__glob: StrIter

    requires: t.Union[StrIter, libdnf5.rpm.ReldepList, libdnf5.rpm.PackageSet]
    requires__eq: t.Union[StrIter, libdnf5.rpm.ReldepList, libdnf5.rpm.PackageSet]
    requires__neq: t.Union[StrIter, libdnf5.rpm.ReldepList, libdnf5.rpm.PackageSet]
    requires__glob: StrIter

    conflicts: t.Union[StrIter, libdnf5.rpm.ReldepList, libdnf5.rpm.PackageSet]
    conflicts__eq: t.Union[StrIter, libdnf5.rpm.ReldepList, libdnf5.rpm.PackageSet]
    conflicts__neq: t.Union[StrIter, libdnf5.rpm.ReldepList, libdnf5.rpm.PackageSet]
    conflicts__glob: StrIter

    obsoletes: t.Union[StrIter, libdnf5.rpm.ReldepList, libdnf5.rpm.PackageSet]
    obsoletes__eq: t.Union[StrIter, libdnf5.rpm.ReldepList, libdnf5.rpm.PackageSet]
    obsoletes__neq: t.Union[StrIter, libdnf5.rpm.ReldepList, libdnf5.rpm.PackageSet]
    obsoletes__glob: StrIter

    recommends: t.Union[StrIter, libdnf5.rpm.ReldepList, libdnf5.rpm.PackageSet]
    recommends__eq: t.Union[StrIter, libdnf5.rpm.ReldepList, libdnf5.rpm.PackageSet]
    recommends__neq: t.Union[StrIter, libdnf5.rpm.ReldepList, libdnf5.rpm.PackageSet]
    recommends__glob: StrIter

    suggests: t.Union[StrIter, libdnf5.rpm.ReldepList, libdnf5.rpm.PackageSet]
    suggests__eq: t.Union[StrIter, libdnf5.rpm.ReldepList, libdnf5.rpm.PackageSet]
    suggests__neq: t.Union[StrIter, libdnf5.rpm.ReldepList, libdnf5.rpm.PackageSet]
    suggests__glob: StrIter

    enhances: t.Union[StrIter, libdnf5.rpm.ReldepList, libdnf5.rpm.PackageSet]
    enhances__eq: t.Union[StrIter, libdnf5.rpm.ReldepList, libdnf5.rpm.PackageSet]
    enhances__neq: t.Union[StrIter, libdnf5.rpm.ReldepList, libdnf5.rpm.PackageSet]
    enhances__glob: StrIter

    supplements: t.Union[StrIter, libdnf5.rpm.ReldepList, libdnf5.rpm.PackageSet]
    supplements__eq: t.Union[StrIter, libdnf5.rpm.ReldepList, libdnf5.rpm.PackageSet]
    supplements__neq: t.Union[StrIter, libdnf5.rpm.ReldepList, libdnf5.rpm.PackageSet]
    supplements__glob: StrIter

    file: StrIter
    file__eq: StrIter
    file__neq: StrIter
    file__glob: StrIter
    file__contains: StrIter

    latest: int
    latest_per_arch: int

    downgrades: bool

    empty: bool

    reponame: StrIter
    reponame__eq: StrIter
    reponame__neq: StrIter
    reponame__glob: StrIter
    reponame__contains: StrIter

    pkg: Iterable[libdnf5.rpm.Package]
    pkg__eq: Iterable[libdnf5.rpm.Package]
    pkg__neq: Iterable[libdnf5.rpm.Package]


class BaseMaker(BaseMakerBase):
    """
    Create a Base object and load repos
    """

    base: libdnf5.base.Base

    def __init__(
        self,
        base: libdnf5.base.Base | None = None,
        *,
        initialized: bool = False,
        config_loaded: bool = False,
    ) -> None:
        """
        Initialize and configure the base object.
        :param base: Pass in a :class:`libdnf.base.Base object` to configure
        instead of creating a new one.
        :param initialized: Set to True if base.setup() has already been
        called. Only applies when `base` is passed.
        :param config_loaded: Set to True if base.load_config_from_file()
        has already been called. Only applies when `base` is passed.
        """
        self.base = base or libdnf5.base.Base()
        self.initialized = initialized if base else False
        if not base or not config_loaded:
            self.base.load_config_from_file()

    def setup(self) -> None:
        if not self.initialized:
            self.base.setup()
            self.initialized = True

    # Not part of the BaseMakerBase interface
    @property
    def config(self) -> libdnf5.config.ConfigMain:
        return self.base.get_config()

    # Not part of the BaseMakerBase interface
    @property
    def vars(self) -> libdnf5.conf.Vars:
        return self.base.get_vars()

    # Not part of the BaseMakerBase interface
    def _set(self, config, key: str, value: t.Any) -> None:
        self._get_option(config, key).set(Priority_RUNTIME, value)

    def set(self, key: str, value: t.Any) -> None:
        # if self.initialized:
        #     raise RuntimeError("The base object has already been initialized")
        LOG.debug("Setting config option %s=%r", key, value)
        self._set(self.config, key, value)

    def set_var(self, key: str, value: t.Any) -> None:
        self.vars.set(key, value)

    # Not part of the BaseMakerBase interface
    @property
    def rs(self) -> libdnf5.repo.RepoSackWeakPtr:
        self.setup()
        return self.base.get_repo_sack()

    def fill_sack(
        self,
        *,
        from_cache: bool = False,
        load_system_repo: bool = False,
    ) -> libdnf5.base.Base:
        """
        Fill the sack and returns the Base object.
        The repository configuration shouldn't be manipulated after this.

        Note that the `_cachedir` arg is private and subject to removal.
        """
        if from_cache:
            raise NotImplementedError
        self.rs.update_and_load_enabled_repos(load_system_repo)
        return self.base

    def read_system_repos(self, disable: bool = True) -> None:
        """
        Load system repositories into the base object.
        By default, they are all disabled even if 'enabled=1' is in the
        repository configuration.
        """
        self.rs.create_repos_from_system_configuration()
        if disable:
            repoq = libdnf5.repo.RepoQuery(self.base)
            repoq.filter_enabled(True)
            for repo in repoq:
                repo.disable()

    def enable_repos(self, repos: Collection[str]) -> None:
        """
        Enable a list of repositories by their repoid.
        Raise a ValueError if the repoid is not in `self.base`'s configuration.
        """
        for repo in repos:
            self.enable_repo(repo)

    def enable_repo(self, repo: str) -> None:
        """
        Enable a repo by its id.
        Raise a ValueError if the repoid is not in `self.base`'s configuration.
        """
        repoq = libdnf5.repo.RepoQuery(self.base)
        repoq.filter_id(repo, libdnf5.common.QueryCmp_GLOB)
        if not repoq:
            raise ValueError(f"{repo} repo definition was not found.")
        for repo in repoq:
            repo.enable()  # type: ignore

    def disable_repo(self, repo: str, ignore_missing: bool = True) -> None:
        """
        Disable a repo by its id.
        Raise a ValueError if the repoid is not in `self.base`'s configuration
        when ignore_missing is False.
        """
        repoq = libdnf5.repo.RepoQuery(self.base)
        repoq.filter_id(repo, libdnf5.common.QueryCmp_GLOB)
        if not ignore_missing and not repoq:
            raise ValueError(f"{repo} repo definition was not found.")
        for repo in repoq:
            repo.disable()  # type: ignore

    def read_repofile(self, file: StrPath) -> None:
        """
        Load repositories from a repo file
        """
        self.rs.create_repos_from_file(str(file))

    def _get_option(self, config: libdnf5.conf.Config, key: str) -> libdnf5.conf.Option:
        """
        Get an Option object from a libdnf5 Config object.
        Maintains compatability with dnf5 versions before
        https://github.com/rpm-software-management/dnf5/pull/327
        """
        LOG.debug("Getting option %s", key)
        # dnf5 > 5.0.7
        if option := getattr(config, f"get_{key}_option", None):
            LOG.debug(f"option = get_{key}_options")
            return option()
        # dnf5 <= 5.0.7
        # TODO: Add warning and deprecate
        elif option := getattr(config, key, None):
            return option()
        else:
            raise ValueError(f"{key!r} is not a valid option.")

    # This is private for now
    def _read_repofile_new(self, file: StrPath, ensure_enabled: bool = False) -> None:
        """
        Load repositories from a repo file if they're not already in the
        configuration.
        """
        parser = libdnf5.conf.ConfigParser()
        parser.read(str(file))
        sections = parser.get_data()
        for name, _ in sections:
            expanded_id = self.vars.substitute(name)
            if expanded_id == "main":
                LOG.debug("Not reading main section.")
                continue
            if repo := self._get_repo(expanded_id):
                LOG.debug("Not adding %s. It's already in the config.", expanded_id)
            else:
                LOG.debug("Adding %s from %s", expanded_id, file)
                repo = self.rs.create_repo(expanded_id)
                repo_config: libdnf5.repo.ConfigRepo = repo.get_config()
                repo_config.load_from_parser(
                    parser,
                    expanded_id,
                    # base.get_vars() returns a WeakPtr, hence the .get()
                    self.base.get_vars().get(),
                    # base.get_logger() returns a WeakPtr, hence the .get()
                    self.base.get_logger().get(),
                    Priority_RUNTIME,
                )
                nameop = self._get_option(repo_config, "name")
                if nameop.get_priority() == libdnf5.conf.Option.Priority_DEFAULT:
                    nameop.set(Priority_RUNTIME, expanded_id)
            if ensure_enabled:
                repo.enable()

    def _get_repo(self, name: str) -> libdnf5.repo.Repo | None:
        """
        Get a repository. Returns None if the repository doesn't exist.
        """
        repoq = libdnf5.repo.RepoQuery(self.base)
        repoq.filter_id([name])
        return next(iter(repoq), None)

    def load_filelists(self) -> None:
        LOG.debug("Loading filelists")
        option = self._get_option(self.config, "optional_metadata_types")
        func = option.add_item
        if "priority" in inspect.signature(func).parameters:
            func(Priority_RUNTIME, libdnf5.conf.METADATA_TYPE_FILELISTS)
        else:
            func(libdnf5.conf.METADATA_TYPE_FILELISTS)

    def create_repo(self, repoid: str, **kwargs) -> None:
        """
        Add a Repo object to the repo sack and configure it.
        :param kwargs: key-values options that should be set on the Repo object
                       values (like $basearch) will be substituted automatically.
        """
        repo = self.rs.create_repo(repoid)
        config = repo.get_config()
        for key, value in kwargs.items():
            value = self._substitute(value)
            self._set(config, key, value)

    # https://github.com/rpm-software-management/dnf5/issues/306
    # https://github.com/rpm-software-management/dnf/blob/276ade38231f3f4120f442e5c3d214f37b66379b/dnf/repodict.py#L73
    def _substitute(self, values):
        if isinstance(values, str):
            return self.vars.substitute(values)
        if isinstance(values, (list, tuple)):
            new = []
            for value in values:
                if isinstance(value, str):
                    new.append(self.vars.substitute(value))
                else:
                    new.append(value)
            return new
        return values

    @property
    def backend(self) -> BackendMod:
        return sys.modules[__name__]

    def repolist(self, enabled: bool | None = None) -> list[str]:
        repoq = libdnf5.repo.RepoQuery(self.base)
        if enabled is not None:
            repoq.filter_enabled(enabled)
        return [r.get_id() for r in repoq]


@functools.total_ordering
class Package(libdnf5.rpm.Package):
    DEBUGINFO_SUFFIX = "-debuginfo"
    DEBUGSOURCE_SUFFIX = "-debugsource"
    """
    libdnf5.rpm.Package subclass with strong dnf.package.Package compatability
    """

    def __hash__(self) -> int:
        return hash(self.get_id().id)

    def __gt__(self, other) -> bool:
        if not isinstance(other, libdnf5.rpm.Package):
            raise TypeError
        if self.name != other.name:
            return self.name > other.name
        evrcmp = libdnf5.rpm.rpmvercmp(self.get_evr(), other.get_evr())
        if evrcmp != 0:
            return evrcmp > 0
        return self.get_arch() > other.get_arch()

    def __lt__(self, other) -> bool:
        return not (self > other)

    @property
    def name(self) -> str:
        return self.get_name()

    @property
    def arch(self) -> str:
        return self.get_arch()

    @property
    def a(self) -> str:
        return self.get_arch()

    @property
    def epoch(self) -> int:
        return int(self.get_epoch())

    @property
    def e(self) -> int:
        return self.epoch

    @property
    def version(self) -> str:
        return self.get_version()

    @property
    def v(self) -> str:
        return self.version

    @property
    def release(self) -> str:
        return self.get_release()

    @property
    def r(self) -> str:
        return self.release

    @property
    def from_repo(self) -> str:
        return self.get_from_repo_id()

    @property
    def evr(self) -> str:
        return self.get_evr()

    @property
    def debug_name(self) -> str:
        # Taken from dnf.package.Package
        # Copyright (C) 2012-2016 Red Hat, Inc.
        # SPDX-License-Identifier: GPL-2.0-or-later
        """
        Returns name of the debuginfo package for this package.
        If this package is a debuginfo package, returns its name.
        If this package is a debugsource package, returns the debuginfo package
        for the base package.
        e.g. kernel-PAE -> kernel-PAE-debuginfo
        """
        if self.name.endswith(self.DEBUGINFO_SUFFIX):
            return self.name

        name = self.name
        if self.name.endswith(self.DEBUGSOURCE_SUFFIX):
            name = name[: -len(self.DEBUGSOURCE_SUFFIX)]

        return name + self.DEBUGINFO_SUFFIX

    @property
    def source_name(self) -> t.Optional[str]:
        # def source_name(self) -> str:
        return None if self.arch == "src" else self.get_source_name()
        # return self.get_source_name()

    @property
    def source_debug_name(self) -> str:
        source_name = self.name if self.arch == "src" else self.source_name
        return source_name + self.DEBUGSOURCE_SUFFIX  # type: ignore[operator]

    @property
    def installtime(self) -> int:
        return self.get_install_time()

    @property
    def buildtime(self) -> int:
        return self.get_build_time()

    @property
    def size(self) -> int:
        return self.get_package_size()

    @property
    def downloadsize(self) -> int:
        return self.get_package_size()

    @property
    def installsize(self) -> int:
        return self.get_install_size()

    @property
    def provides(self) -> Iterable[Reldep5]:
        return self.get_provides()

    @property
    def requires(self) -> Iterable[Reldep5]:
        return self.get_requires()

    @property
    def recommends(self) -> Iterable[Reldep5]:
        return self.get_recommends()

    @property
    def suggests(self) -> Iterable[Reldep5]:
        return self.get_suggests()

    @property
    def supplements(self) -> Iterable[Reldep5]:
        return self.get_supplements()

    @property
    def enhances(self) -> Iterable[Reldep5]:
        return self.get_enhances()

    @property
    def obsoletes(self) -> Iterable[Reldep5]:
        return self.get_obsoletes()

    @property
    def conflicts(self) -> Iterable[Reldep5]:
        return self.get_conflicts()

    @property
    def sourcerpm(self) -> t.Optional[str]:
        return self.get_sourcerpm() or None

    @property
    def description(self) -> str:
        return self.get_description()

    @property
    def summary(self) -> str:
        return self.get_summary()

    @property
    def license(self) -> str:
        return self.get_license()

    @property
    def url(self) -> str:
        return self.get_url()

    @property
    def reason(self) -> str:
        reason = self.get_reason()
        reason_str = libdnf5.transaction.transaction_item_reason_to_string(reason)
        return reason_str

    @property
    def files(self) -> Iterable[str]:
        return self.get_files()

    @property
    def reponame(self) -> str:
        return self.get_repo_id()

    @property
    def repoid(self) -> str:
        return self.get_repo_id()

    @property
    def vendor(self) -> str:
        return self.get_vendor()

    @property
    def packager(self) -> str:
        return self.get_packager()

    @property
    def location(self) -> str:
        return self.get_location()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}<{str(self)}>"

    def __str__(self) -> str:
        return self.get_nevra()


libdnf5._rpm.Package_swigregister(Package)


class Reldep5(libdnf5.rpm.Reldep):
    """
    Subclass of libdnf5.rpm.Reldep with a useful __str__() method
    """

    def __str__(self) -> str:
        return self.to_string()


libdnf5._rpm.Reldep_swigregister(Reldep5)


class PackageQuery(libdnf5.rpm.PackageQuery):
    __rq__: Repoquery

    """
    Subclass of libdnf5.rpm.PackageQuery with hawkey.Query compatability
    """

    def _filter(  # type: ignore[override]
        self,
        **kwargs: Unpack[_QueryFilterKwargs],
    ) -> None:
        if not kwargs:
            return None
        filter_mapping = {
            "latest": "latest_evr",
            "latest_per_arch": "latest_evr",
            "reponame": "repo_id",
        }
        comp_mapping: dict[str, int] = {
            "eq": libdnf5.common.QueryCmp_EQ,
            "neq": libdnf5.common.QueryCmp_NEQ,
            "glob": libdnf5.common.QueryCmp_GLOB,
            "contains": libdnf5.common.QueryCmp_NOT_CONTAINS,
            "gt": libdnf5.common.QueryCmp_GT,
            "gte": libdnf5.common.QueryCmp_GTE,
            "lt": libdnf5.common.QueryCmp_LT,
            "lte": libdnf5.common.QueryCmp_LTE,
        }
        invalid = []
        for key in kwargs:
            if key not in _QueryFilterKwargs.__annotations__:
                invalid.append(key)
        if invalid:
            raise TypeError(f"Invalid keyword arguments: {invalid}")
        if kwargs.pop("empty", None):
            self.clear()
            return None
        if kwargs.pop("downgrades", None):
            self.filter_downgrades()
        for key, value in kwargs.items():
            split = key.rsplit("__", 1)
            name = "filter_" + filter_mapping.get(split[0], split[0])
            args = [_convert_value(key, value)]
            if len(split) == 2:
                args.append(comp_mapping[split[1]])
            getattr(self, name)(*args)

    def filterm(  # type: ignore[override]
        self,
        **kwargs: Unpack[_QueryFilterKwargs],
    ) -> PackageQuery:
        self._filter(**kwargs)
        return self

    def filter(  # type: ignore[override]
        self,
        **kwargs: Unpack[_QueryFilterKwargs],
    ) -> PackageQuery:
        self._filter(**kwargs)
        return self

    def __len__(self) -> int:
        return self.size()

    def union(self, other: PackageQuery) -> PackageQuery:
        self.update(other)
        return self

    def __ior__(self, other: PackageQuery) -> PackageQuery:
        return self.union(other)

    _pkg_comps: TypeAlias = (
        "t.Union[libdnf5.common.QueryCmp_EQ, libdnf5.common.QueryCmp_NEQ]"
    )

    def filter_pkg(
        self,
        pkgs: Iterable[libdnf5.rpm.Package],
        comp: _pkg_comps = libdnf5.common.QueryCmp_EQ,
        /,
    ):
        if isinstance(pkgs, (libdnf5.rpm.PackageSet, libdnf5.rpm.PackageQuery)):
            newquery = pkgs
        else:
            if hasattr(self, "__rq__"):
                base = self.__rq__.base
            else:
                base = self.get_base()
            newquery = libdnf5.rpm.PackageSet(base)
            newquery.clear()
            for p in pkgs:
                newquery.add(p)
        if comp == libdnf5.common.QueryCmp_EQ:
            self.intersection(newquery)
        elif comp == libdnf5.common.QueryCmp_NEQ:
            self.difference(newquery)
        else:
            raise ValueError()

    def __str__(self) -> str:
        return f"{self.__class__.__name__}<{tuple(self)}>"


_ValT = t.TypeVar("_ValT")


def _convert_value(key: str, value: _ValT) -> t.Union[list[_ValT], _ValT]:
    """
    Overkill code to normalize single strings or ints that libdnf5 requires to
    be a list based on type annotations.
    """
    d_annotations = t.get_type_hints(_QueryFilterKwargs)
    annotation = d_annotations[key]
    if t.get_origin(annotation) is not t.Union or not any(
        isinstance(value, typ) for typ in CONVERT_TO_LIST
    ):
        return value
    union_types = t.get_args(annotation)
    pairs = [(t.get_origin(arg), t.get_args(arg)) for arg in union_types]
    for origin, sargs in pairs:
        assert isinstance(origin, type)
        if not issubclass(origin, list):
            continue
        if t.get_origin(sargs[0]) is t.Union:
            pairs = [(t.get_origin(arg), t.get_args(arg)) for arg in union_types]
            for origin, sargs in pairs:
                assert isinstance(origin, type)
                if issubclass(origin, list) and isinstance(value, sargs[0]):
                    return [value]
        if isinstance(value, sargs[0]):
            return [value]
    return value


class Repoquery(RepoqueryBase):
    def __init__(self, base: libdnf5.base.Base) -> None:
        self.base: libdnf5.base.Base = base

    @property
    def base_arches(self) -> set[str]:
        base_vars = self.base.get_vars()
        return {base_vars.get_value("arch"), base_vars.get_value("basearch")}

    def _query(self) -> PackageQuery:
        obj = PackageQuery(self.base)
        obj.__rq__ = self
        return obj

    def query(
        self, *, arch: str | Iterable[str] | None = None, **kwargs
    ) -> PackageQuery:
        return t.cast(PackageQuery, super().query(arch=arch, **kwargs))

    def resolve_pkg_specs(
        self,
        specs: Collection[str],
        resolve: bool = False,
        latest: int | None = None,
        with_src: bool = True,
    ) -> PackageQuery:
        settings = libdnf5.base.ResolveSpecSettings()
        settings.with_filenames = resolve
        settings.with_provides = resolve

        r_query = self.query(empty=True)
        for spec in specs:
            query = self._query()
            query.resolve_pkg_spec(spec, settings, with_src)
            r_query.union(query)
        if resolve:
            r_query = r_query.union(self.query(provides=specs))
        filter_latest(r_query, latest)
        return r_query

    @property
    def backend(self) -> BackendMod:
        return sys.modules[__name__]


def _dnf_getreleasever() -> str:
    # This is taken from dnf and slightly modified
    #
    # SPDX-License-Identifier: GPL-2.0-or-later
    # Copyright (C) 2012-2015  Red Hat, Inc.
    DISTROVERPKG = (
        "system-release(releasever)",
        "system-release",
        "distribution-release(releasever)",
        "distribution-release",
        "redhat-release",
        "suse-release",
    )
    ts = rpm.TransactionSet("/")
    try:
        ts.setVSFlags(~(rpm._RPMVSF_NOSIGNATURES | rpm._RPMVSF_NODIGESTS))
        for distroverpkg in map(lambda p: p.encode("utf-8"), DISTROVERPKG):
            idx = ts.dbMatch("provides", distroverpkg)
            if not len(idx):
                continue
            try:
                hdr = next(idx)
            except StopIteration:
                raise RuntimeError(
                    "Error: rpmdb failed to list provides. Try: rpm --rebuilddb"
                ) from None
            releasever = hdr["version"]
            try:
                try:
                    # header returns bytes -> look for bytes
                    # it may fail because rpm returns a decoded string since 10 Apr 2019
                    off = hdr[rpm.RPMTAG_PROVIDENAME].index(distroverpkg)
                except ValueError:
                    # header returns a string -> look for a string
                    off = hdr[rpm.RPMTAG_PROVIDENAME].index(distroverpkg.decode("utf8"))
                flag = hdr[rpm.RPMTAG_PROVIDEFLAGS][off]
                ver = hdr[rpm.RPMTAG_PROVIDEVERSION][off]
                if flag == rpm.RPMSENSE_EQUAL and ver:
                    if hdr["name"] not in (distroverpkg, distroverpkg.decode("utf8")):
                        # override the package version
                        releasever = ver
            except (ValueError, KeyError, IndexError):
                pass
            if isinstance(releasever, bytes):
                releasever = releasever.decode("utf-8")
            return releasever
        return ""
    finally:
        ts.closeDB()


@functools.cache
def get_releasever() -> str:
    """
    Return the system releasever
    """
    # libdnf5 > 5.0.10
    if hasattr(libdnf5.conf.Vars, "detect_release"):
        base = libdnf5.base.Base()
        return libdnf5.conf.Vars.detect_release(base.get_weak_ptr(), "/").get()
    # Fall back to our copy of dnf4's code
    else:
        return _dnf_getreleasever()


RepoError = RuntimeError

__all__ = (
    "BACKEND",
    "BaseMaker",
    "Package",
    "PackageQuery",
    "Repoquery",
    "RepoError",
    "get_releasever",
    #
    "libdnf5",
)



# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import argparse
import collections.abc as cabc
import logging
import typing as t

from fedrq._utils import filter_latest, get_source_name
from fedrq.backends.base import PackageCompat, PackageQueryCompat
from fedrq.cli.base import Command
from fedrq.cli.formatters import DefaultFormatters, Formatter

logger = logging.getLogger(__name__)


class BreakdownFormatter(Formatter):
    MULTILINE = True

    def format_line(self, package: PackageCompat) -> str:
        raise NotImplementedError

    def format(self, packages: cabc.Iterable[PackageCompat]) -> cabc.Iterable[str]:
        runtime = []
        buildtime = []
        for p in packages:
            if p.arch == "src":
                buildtime.append(p)
            else:
                runtime.append(p)
        if runtime:
            yield "Runtime:"
            for p in sorted(runtime):
                yield p.name
            yield f"    {len(runtime)} total runtime dependencies"
            if buildtime:
                yield ""
        if buildtime:
            yield "Buildtime:"
            for p in sorted(buildtime):
                yield p.name
            yield f"    {len(buildtime)} total buildtime dependencies"
        yield ""
        yield "All SRPM names:"
        yield from (all_pkgs := sorted({get_source_name(pkg) for pkg in packages}))
        yield f"    {len(all_pkgs)} total SRPMs"


WhatFormatters = DefaultFormatters | dict(breakdown=BreakdownFormatter)


class WhatCommand(Command):
    formatters = WhatFormatters
    _exclude_subpackages_opt: bool = False
    _operator: str
    operator: str

    def __init__(self, args: argparse.Namespace) -> None:
        super().__init__(args)
        self.v_default()

    @classmethod
    def make_parser(
        cls,
        parser_func: cabc.Callable = argparse.ArgumentParser,
        *,
        add_help: bool = False,
        **kwargs,
    ) -> argparse.ArgumentParser:
        pargs = dict(description=cls.__doc__, parents=[cls.parent_parser()], **kwargs)
        if add_help:
            pargs["help"] = f"Find reverse {cls.operator.title()} of a list of packages"
        parser = parser_func(**pargs)

        _rp_help = f"""
        Resolve the correct Package when given a virtual Provide. For instance,
        /usr/bin/yt-dlp would resolve to yt-dlp, and then any package that
        {cls.operator} python3dist(yt-dlp) would also be included.
        """
        resolve_group = parser.add_mutually_exclusive_group()
        resolve_group.add_argument(
            "-P", "--resolve-packages", action="store_true", help=_rp_help
        )
        resolve_group.add_argument(
            "-E",
            "--exact",
            action="store_true",
            help="This is the opposite extreme to --resolve-packages. "
            "E.g., yt-dlp would not match python3dist(yt-dlp) like it does by default.",
        )

        arch_group = parser.add_mutually_exclusive_group()
        arch_group.add_argument(
            "-A",
            "--arch",
            help=f"After finding the packages that {cls._operator} NAMES, "
            "filter out the resulting packages that don't match ARCH",
        )
        arch_group.add_argument(
            "-S",
            "--notsrc",
            dest="arch",
            action="store_const",
            const="notsrc",
            help="This includes all binary RPMs. Multilib is excluded on x86_64. "
            "Equivalent to --arch=notsrc",
        )
        arch_group.add_argument(
            "-s",
            "--src",
            dest="arch",
            action="store_const",
            const="src",
            help="Query for BuildRequires of NAME. "
            "This is equivalent to --arch=src.",
        )
        if cls._exclude_subpackages_opt:
            parser.add_argument("-X", "--exclude-subpackages", action="store_true")
        return parser

    def exclude_subpackages(self, rpms: t.Optional[PackageQueryCompat]) -> None:
        import re

        rpms = rpms or self.rq.resolve_pkg_specs(self.args.names, resolve=True)
        brpms = self.rq.query(pkg=rpms, arch__neq="src")
        srpms = self.rq.query(pkg=rpms, arch="src")

        brpm_sourcerpms = [
            re.sub(r"\.rpm$", "", t.cast(str, pkg.sourcerpm)) for pkg in brpms
        ]
        brpm_srpm_query = self.rq.resolve_pkg_specs(brpm_sourcerpms)
        subpackages = self.rq.get_subpackages(brpm_srpm_query.union(srpms))
        self.query.filterm(pkg__neq=subpackages)
        return None

    def run(self) -> None:
        self.query = self.rq.query(empty=True)
        # Resolve self.args.names into Package objs.
        # This makes it so packages that depend on virtual Provides of the
        # names are included.
        if not self.args.exact:
            resolved_packages = self.rq.resolve_pkg_specs(
                self.args.names, self.args.resolve_packages, with_src=False
            )
            self._logq(resolved_packages, "resolved_packages")
            operator_kwargs = {self.operator: resolved_packages}
            rp_rdeps = self.rq.query(**operator_kwargs, arch=self.args.arch)
            self._logq(rp_rdeps, "rp_rdeps")

            self.query = self.query.union(rp_rdeps)

        operator_kwargs = {f"{self.operator}__glob": self.args.names}
        glob_rdeps = self.rq.query(**operator_kwargs, arch=self.args.arch)
        self._logq(glob_rdeps, "glob_rdeps")

        self.query = self.query.union(glob_rdeps)
        filter_latest(self.query, self.args.latest)

        if getattr(self.args, "exclude_subpackages", None):
            self.exclude_subpackages(
                resolved_packages if self.args.resolve_packages else None
            )

        for p in self.format():
            print(p)


class Whatrequires(WhatCommand):
    """
    By default, fedrq-whatrequires takes one or more valid package names. Then,
    it finds the packages' reverse dependencies, including dependents of their
    virtual Provides. Use the options below to modify fedrq-whatrequires exact
    search strategy.
    """

    _operator = "Require"
    operator = "requires"
    _exclude_subpackages_opt = True


class Whatrecommends(WhatCommand):
    """
    By default, fedrq-whatrecommends takes one or more valid package names. Then,
    it finds the packages' reverse dependencies, including dependents of their
    virtual Provides. Use the options below to modify fedrq-whatrecommends exact
    search strategy.
    """

    _operator = "Recommend"
    operator = "recommend"
    _exclude_subpackages_opt = True


class Whatsuggests(WhatCommand):
    """
    By default, fedrq-whatsuggests takes one or more valid package names. Then,
    it finds the packages' reverse dependencies, including dependents of their
    virtual Provides. Use the options below to modify fedrq-whatsuggests exact
    search strategy.
    """

    _operator = "Suggest"
    operator = "suggests"
    _exclude_subpackages_opt = True


class Whatsupplements(WhatCommand):
    """
    By default, fedrq-whatsupplements takes one or more valid package names. Then,
    it finds the packages' reverse dependencies, including dependents of their
    virtual Provides. Use the options below to modify fedrq-whatsupplements exact
    search strategy.
    """

    _operator = "Supplement"
    operator = "supplements"
    _exclude_subpackages_opt = False


class Whatenhances(WhatCommand):
    """
    By default, fedrq-whatenhances takes one or more valid package names. Then,
    it finds the packages' reverse dependencies, including dependents of their
    virtual Provides. Use the options below to modify fedrq-whatenhances exact
    search strategy.
    """

    _operator = "Enhance"
    operator = "enhances"
    _exclude_subpackages_opt = False


class WhatrequiresSrc(WhatCommand):
    """
    By default, fedrq whatrequires-src takes one or more valid source package
    names. Then, it finds the reverse dependencies of the binary RPMs that they
    produce. Use the options below to modify fedrq whatrequires-src's exact
    search strategy.
    This command is a shortcut for `fedrq whatrequires $(fedrq subpkgs ...)`.
    """

    _exclude_subpackages_opt = True
    _operator = "Require"
    operator = "requires"

    @classmethod
    def make_parser(
        cls,
        parser_func: cabc.Callable = argparse.ArgumentParser,
        *,
        add_help: bool = False,
        **kwargs,
    ) -> argparse.ArgumentParser:
        kwargs.update(dict(description=cls.__doc__, parents=[cls.parent_parser()]))
        if add_help:
            _help = (
                f"Find reverse {cls.operator.title()} of a list of"
                " source packages' subpackages."
            )
            kwargs["help"] = _help
        parser = parser_func(**kwargs)
        parser.add_argument(
            "-X",
            "--exclude-subpackages",
            action="store_true",
        )
        arch_group = parser.add_mutually_exclusive_group()
        arch_group.add_argument(
            "-A",
            "--arch",
            help=f"After finding the packages that {cls._operator} NAMES' subpackages, "
            "filter out the resulting packages that don't match ARCH",
        )
        arch_group.add_argument(
            "-S",
            "--notsrc",
            dest="arch",
            action="store_const",
            const="notsrc",
            help="This includes all binary RPMs. Multilib is excluded on x86_64. "
            "Equivalent to --arch=notsrc",
        )
        arch_group.add_argument(
            "-s",
            "--src",
            dest="arch",
            action="store_const",
            const="src",
            help="This is equivalent to --arch=src.",
        )
        return parser

    def run(self) -> None:
        srpms = self.rq.resolve_pkg_specs(self.args.names, with_src=True).filterm(
            arch="src"
        )
        subpackages = self.rq.get_subpackages(srpms)
        qkwargs = {
            self.operator: subpackages,
            "arch": self.args.arch,
            "latest": self.args.latest,
        }
        if self.args.exclude_subpackages:
            qkwargs["pkg__neq"] = subpackages
        self.query = self.rq.query(**qkwargs)
        for p in self.format():
            print(p)


class Whatobsoletes(WhatCommand):
    """
    By default, fedrq-whatobsoletes takes one or more valid package names.
    Then, it finds the packages' reverse dependencies, including dependents of
    their virtual Provides. Use the options below to modify fedrq-whatobsoletes
    # exact search strategy.
    """

    _exclude_subpackages_opt: bool = False
    _operator = "Obsolete"
    operator = "obsoletes"



# Copyright (C) 2023 Maxwell G <gotmax@e.email>
#
# SPDX-License-Identifier: GPL-2.0-or-later

import pytest


@pytest.mark.parametrize(
    "args",
    [
        pytest.param(["packageb"]),
        pytest.param(["packageb.src"]),
    ],
)
def test_whatrequires_src(run_command2, args):
    stdout, stderr = run_command2(["wrsrc", "packageb", "-F", "source"], False)
    assert not stderr
    assert stdout == ["packagea", "packageb"]


def test_whatrequires_src_exclude(run_command2):
    stdout, stderr = run_command2(
        ["whatrequires-src", "packageb", "-F", "source", "-X"], False
    )
    assert not stderr
    assert stdout == ["packagea"]



# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later

import os
import subprocess
from pathlib import Path
from shutil import rmtree

import pytest

import fedrq.cli
from fedrq import config as rqconfig
from fedrq.backends import get_default_backend

TEST_DATA = Path(__file__).parent.resolve() / "test_data"

TEST_REPO_1 = f"""
[testrepo1]
name = testrepo1
baseurl = file://{TEST_DATA / 'repos' / 'repo1' / 'repo'}/
gpgcheck = False
"""

TEST_CONFIG_1 = """
default_branch = "tester"

[releases.testrepo1]
matcher = "^(tester)$"
defs.base = ["testrepo1"]
defpaths = ["testrepo1.repo"]
system_repos = false
"""

# @pytest.fixture(scope="session", autouse=True)


@pytest.fixture(scope="session", autouse=True)
def clear_cache():
    path = rqconfig.get_smartcache_basedir() / "tester"
    rmtree(path, ignore_errors=True)
    try:
        yield
    finally:
        rmtree(path, ignore_errors=True)


@pytest.fixture(scope="session", autouse=True)
def default_backend():
    backend = os.environ.get("FEDRQ_BACKEND")
    return get_default_backend(backend, bool(backend))


@pytest.fixture
def temp_smartcache(monkeypatch, tmp_path):
    path = tmp_path / "smartcache"
    path.mkdir()
    monkeypatch.setenv("XDG_CACHE_HOME", str(path))
    return path


@pytest.fixture
def data_path():
    return TEST_DATA


@pytest.fixture(scope="session", autouse=True)
def repo_test_repo():
    assert TEST_DATA.is_dir()
    assert (buildsh := TEST_DATA / "build.sh").exists()
    subprocess.run(("bash", buildsh), check=True)


@pytest.fixture(autouse=True)
def clear_config(monkeypatch):
    monkeypatch.setattr(rqconfig, "CONFIG_DIRS", ())


@pytest.fixture
def patch_config_dirs(tmp_path, monkeypatch):
    config_dirs = (tmp_path / "custom", tmp_path / "global")
    for d in config_dirs:
        d.mkdir()

    conf = config_dirs[1]
    conf.joinpath("test.toml").write_text(TEST_CONFIG_1)

    repo_dir = conf.joinpath("repos")
    repo_dir.mkdir()
    repo_dir.joinpath("testrepo1.repo").write_text(TEST_REPO_1)

    monkeypatch.setattr(rqconfig, "CONFIG_DIRS", config_dirs)
    return config_dirs[0]


@pytest.fixture
def repo_test_rq(patch_config_dirs):
    return rqconfig.get_config(load_filelists="always").get_rq("tester", "base")


@pytest.fixture(scope="session")
def target_cpu():
    macro = subprocess.run(
        ["rpm", "-E", "%_target_cpu"], text=True, capture_output=True, check=True
    ).stdout.strip()
    assert macro != "%{_target_cpu}"
    return macro


@pytest.fixture
def runs(capsys):
    def runner(args, return_stdout=True):
        fedrq.cli.main(args)
        stdout, stderr = capsys.readouterr()
        result = [stdout.splitlines(), stderr.splitlines()]
        if return_stdout:
            result.append(stdout)
        return result

    return runner


@pytest.fixture
def run_command2(runs, patch_config_dirs):
    return runs



# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later

import pytest

import fedrq.cli

YT_DLP_SUPKGS = {
    "yt-dlp-bash-completion",
    "yt-dlp-zsh-completion",
    "yt-dlp-fish-completion",
}

ARGS = (
    (["yt-dlp"]),
    (["-P", "python3dist(yt-dlp)"]),
    (["yt-dlp.noarch"]),
)


@pytest.mark.no_rpm_mock
@pytest.mark.parametrize("args", ARGS)
def test_whatrequires_exclude_subpackages_f37(capsys, args):
    fedrq.cli.main(["whatrequires", "-b", "f37", "--sc", "-X", "-Fname", *args])
    stdout, stderr = capsys.readouterr()
    stdout_lines = set(stdout.splitlines())
    assert not (stdout_lines & YT_DLP_SUPKGS)
    assert "celluloid" in stdout_lines
    assert not stderr


@pytest.mark.no_rpm_mock
@pytest.mark.parametrize("args", ARGS)
def test_whatrequires_not_exclude_subpackages_f37(capsys, args):
    fedrq.cli.main(["whatrequires", "-b", "f37", "--sc", "-Fname", *args])
    stdout, stderr = capsys.readouterr()
    stdout_lines = set(stdout.splitlines())
    assert stdout_lines & YT_DLP_SUPKGS
    assert "celluloid" in stdout_lines
    assert not stderr


@pytest.mark.no_rpm_mock
def test_whatrequires_resolve(capsys):
    """
    Ensure that SRPM names are not considered when resolving Provides
    E.g. python-setuptools should resolve to python3-setuptools.noarch
    (Provides python-setuptools) instead of python-setuptools.src.
    """
    fedrq.cli.main(["whatrequires", "-b", "f37", "-P", "-Fna", "python-setuptools"])
    stdout, stderr = map(lambda f: f.splitlines(), capsys.readouterr())
    # 3700 as of 2023-01-23. Le
    assert len(stdout) > 3000
    assert "python-pip.src" in stdout and "yt-dlp.src" in stdout
    assert not stderr



#!/usr/bin/env python3
# SPDX-License-Identifier: Unlicense
# SPDX-FileCopyrightText: None

# Roughly equivalent to:
#     fedrq whatrequires --arch=noarch -b rawhide -r buildroot bash | grep '^a'

from fedrq.config import get_config, RQConfig
from fedrq.backends.base import RepoqueryBase

# Load config from filesystem and override some options
config: RQConfig = get_config(backend="libdnf5")

# Query the Fedora Rawhide koji buildroot repositories
# This supports any release configuration builtin to fedrq
# or configured on your local system.
rq: RepoqueryBase = config.get_rq("rawhide", "buildroot")

# Get all noarch packages that start with 'a' and depend on bash
query = rq.query(
    name__glob="a*", arch="noarch", requires=rq.query(name="bash", arch="notsrc")
)
# By using sorted(), you'll get (relatively) consistent ordering between backends
for package in sorted(query):
    print(package)



# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later

import pytest

import fedrq.cli

YT_DLP_SUPKGS = [
    "yt-dlp-bash-completion",
    "yt-dlp-fish-completion",
    "yt-dlp-zsh-completion",
]


@pytest.mark.no_rpm_mock
def test_subpkgs_match1(capsys):
    fedrq.cli.main(
        ["subpkgs", "-b", "f37", "yt-dlp", "--match", "*completion", "-F", "name"]
    )
    stdout, stderr = capsys.readouterr()
    assert sorted(stdout.splitlines()) == YT_DLP_SUPKGS
    assert not stderr


@pytest.mark.no_rpm_mock
def test_subpkgs_match2(capsys):
    # Find the source packages that contain a subpackage that match '*-devel'
    fedrq.cli.main(
        [
            "subpkgs",
            "ansible-core",
            "moby-engine",
            "python-pip",
            "yt-dlp",
            "gh",  # no matches
            "-M",
            "*-fish-completion",
            "-M",
            "*-doc",
            "-Fname",
        ]
    )
    stdout, stderr = capsys.readouterr()
    assert stdout.splitlines() == [
        "ansible-core-doc",
        "moby-engine-fish-completion",
        "python-pip-doc",
        "yt-dlp-fish-completion",
    ]
    assert not stderr



# noqa
from __future__ import annotations

import sys

if sys.version_info < (3, 11):
    # if True:
    # SPDX-License-Identifier: PSF-2.0
    # Copyright (c) 2001-2023 Python Software Foundation
    from enum import Enum
    from typing import Any, TypeVar

    _S = TypeVar("_S", bound="StrEnum")

    class StrEnum(str, Enum):
        """
        Enum where members are also (and must be) strings
        """

        def __new__(cls: type[_S], *values: str) -> _S:
            if len(values) > 3:
                raise TypeError("too many arguments for str(): %r" % (values,))
            if len(values) == 1:
                # it must be a string
                if not isinstance(values[0], str):
                    raise TypeError("%r is not a string" % (values[0],))
            if len(values) >= 2:
                # check that encoding argument is a string
                if not isinstance(values[1], str):
                    raise TypeError("encoding must be a string, not %r" % (values[1],))
            if len(values) == 3:
                # check that errors argument is a string
                if not isinstance(values[2], str):
                    raise TypeError("errors must be a string, not %r" % (values[2]))
            value = str(*values)
            member = str.__new__(cls, value)
            member._value_ = value
            return member

        __str__ = str.__str__

        @staticmethod
        def _generate_next_value_(
            name: str, start: int, count: int, last_values: list[Any]
        ) -> str:
            """
            Return the lower-cased version of the member name.
            """
            return name.lower()

else:
    from enum import StrEnum



# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later

from pathlib import Path

try:
    import tomllib
except ImportError:
    import tomli as tomllib

import pytest
import tomli_w

import fedrq.cli


@pytest.fixture
def no_tomli_w(monkeypatch, mocker):
    monkeypatch.setattr(fedrq.cli.base, "HAS_TOMLI_W", False)
    mock = mocker.patch("fedrq.cli.base.tomli_w")
    mock.dump.side_effect = NameError


def test_checkconfig_basic(run_command2):
    out = run_command2(["check-config"])
    assert "No validation errors found!" in out[0]
    assert not out[1]


def test_checkconfig_dump(run_command2, patch_config_dirs):
    defs = {"base": ["testrepo1"]}
    expected = {
        "append_system_repos": False,
        "matcher": "^(tester)$",
        "repo_aliases": {},
        "defpaths": ["testrepo1.repo"],
        "system_repos": False,
        "defs": defs,
        "repo_dirs": list(
            map(
                str,
                (
                    patch_config_dirs / "repos",
                    patch_config_dirs.parent / "global" / "repos",
                ),
            )
        ),
    }

    out = run_command2(["check-config", "--dump"])
    data = tomllib.loads(out[2])

    testrepo1 = data["releases"]["testrepo1"]
    del testrepo1["full_def_paths"]
    assert testrepo1 == expected
    assert not out[1]


def test_checkconfig_dump_error(capsys, no_tomli_w):
    with pytest.raises(SystemExit, match="tomli-w is required for --dump"):
        fedrq.cli.main(["check-config", "--dump"])
    stdout = capsys.readouterr()[0]
    assert not stdout


def test_checkconfig_default_branch_error(capsys, patch_config_dirs):
    dest: Path = patch_config_dirs / "99-default_branch_error.toml"
    data = {"default_branch": "does-not-exist"}
    with dest.open("wb") as fp:
        tomli_w.dump(data, fp)
    with pytest.raises(SystemExit, match="default_branch 'does-not-exist' is invalid"):
        fedrq.cli.main(["check-config"])
    stdout = capsys.readouterr()[0]
    assert stdout == "Validating config...\n"



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
#
# SPDX-License-Identifier: GPL-2.0-or-later

import pytest

import fedrq.config


@pytest.mark.parametrize(
    "args",
    [
        pytest.param("gotmax23/fedrq"),
        pytest.param("gotmax23/fedrq@copr.fedoraproject.org/"),
        pytest.param("gotmax23/fedrq@https://copr.fedoraproject.org/"),
        pytest.param("gotmax23/fedrq@copr.fedoraproject.org"),
        pytest.param("gotmax23/fedrq@https://copr.fedoraproject.org"),
    ],
)
def test_copr_repo(args):
    config = fedrq.config.get_config()
    f37 = config.get_release("f37")
    url = f37._copr_repo(args)
    assert url == "https://copr.fedoraproject.org/coprs/gotmax23/fedrq/repo/fedora-37"


@pytest.mark.parametrize(
    "args",
    [
        pytest.param("@python/python3.12"),
        pytest.param("@python/python3.12@copr.fedoraproject.org/"),
        pytest.param("@python/python3.12@https://copr.fedoraproject.org/"),
        pytest.param("@python/python3.12@copr.fedoraproject.org"),
        pytest.param("@python/python3.12@https://copr.fedoraproject.org"),
    ],
)
def test_copr_repo_group(args):
    config = fedrq.config.get_config()
    rawhide = config.get_release("rawhide")
    url = rawhide._copr_repo(args)
    expected = (
        "https://copr.fedoraproject.org/coprs/g/python/python3.12/repo/fedora-rawhide"
    )
    assert url == expected


def test_copr_custom():
    config = fedrq.config.get_config()
    rawhide = config.get_release("rawhide")
    url = rawhide._copr_repo("@python/python3.12@example.com")
    assert url == "https://example.com/coprs/g/python/python3.12/repo/fedora-rawhide"



# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import argparse
from collections import abc as cabc

from fedrq._utils import filter_latest, mklog
from fedrq.cli import Command


class Pkgs(Command):
    """
    Find the packages that match NAMES.
    NAMES can be package package name globs or NEVRs.
    """

    def __init__(self, args: argparse.Namespace) -> None:
        super().__init__(args)
        self.v_default()

    @classmethod
    def make_parser(
        cls,
        parser_func: cabc.Callable = argparse.ArgumentParser,
        *,
        add_help: bool = False,
        **kwargs,
    ) -> argparse.ArgumentParser:
        kwargs.update(dict(description=Pkgs.__doc__, parents=[cls.parent_parser()]))
        if add_help:
            kwargs["help"] = "Find the packages that match a list of package specs"
        parser = parser_func(**kwargs)

        parser.add_argument(
            "-P",
            "--resolve-packages",
            action="store_true",
            help="Resolve the correct Package when given a virtual Provide."
            " For instance, /usr/bin/yt-dlp would resolve to yt-dlp",
        )

        arch_group = parser.add_mutually_exclusive_group()
        arch_group.add_argument(
            "-A",
            "--arch",
            help="Only include packages that match ARCH",
        )
        arch_group.add_argument(
            "-S",
            "--notsrc",
            dest="arch",
            action="store_const",
            const="notsrc",
            help="This includes all binary RPMs. Multilib is excluded on x86_64. "
            "Equivalent to --arch=notsrc",
        )
        arch_group.add_argument(
            "-s",
            "--src",
            dest="arch",
            action="store_const",
            const="src",
            help="Query for BuildRequires of NAME. "
            "This is equivalent to --arch=src.",
        )
        return parser

    def run(self) -> None:
        flog = mklog(__name__, self.__class__.__name__, "run")
        self.query = self.rq.query(empty=True)
        # flog.debug("self.query = %s", tuple(self.query))

        resolved_packages = self.rq.resolve_pkg_specs(
            self.args.names, self.args.resolve_packages
        )
        self._logq(resolved_packages, "resolved_packages")
        self.query = self.query.union(resolved_packages)

        glob_packages = self.rq.query(name__glob=self.args.names)
        self._logq(glob_packages, "glob_packages")
        self.query = self.query.union(glob_packages)

        self.query = self.rq.arch_filter(self.query, self.args.arch)
        filter_latest(self.query, self.args.latest)
        flog.debug("self.query = %s", tuple(self.query))

        for p in self.format():
            print(p)



# Copyright (C) 2023 Maxwell G <gotmax@e.email>
#
# SPDX-License-Identifier: GPL-2.0-or-later

import pytest

from fedrq.backends.base import BackendMod

# ruff: noqa: E501
TEST_REPO = "https://download.copr.fedorainfracloud.org/results/gotmax23/community.general.copr_integration_tests/fedora-$releasever-$basearch/"
TEST_REPO_KEY = "https://download.copr.fedorainfracloud.org/results/gotmax23/community.general.copr_integration_tests/pubkey.gpg"
TEST_BASEARCH = "x86_64"
TEST_RELEASEVER = "37"


@pytest.mark.no_rpm_mock
def test_bm_create_repo_failure(default_backend: BackendMod) -> None:
    bm = default_backend.BaseMaker()
    bm.sets({}, dict(basearch=TEST_BASEARCH, releasever="nonexistant"))
    bm.create_repo(
        "copr_integration_tests",
        baseurl=[TEST_REPO],
        skip_if_unavailable=False,
    )
    with pytest.raises(default_backend.RepoError):
        default_backend.Repoquery(bm.fill_sack())


@pytest.mark.parametrize(
    "baseurl",
    (
        (TEST_REPO),
        (
            TEST_REPO.replace("$basearch", TEST_BASEARCH).replace(
                "$releasever", TEST_RELEASEVER
            )
        ),
    ),
)
@pytest.mark.no_rpm_mock
def test_bm_create_repo_full(default_backend: BackendMod, baseurl):
    bm = default_backend.BaseMaker()
    bm.sets({}, dict(basearch=TEST_BASEARCH, releasever=TEST_RELEASEVER))
    bm.create_repo(
        "copr_integration_tests",
        baseurl=[TEST_REPO],
        skip_if_unavailable=False,
        type="rpm-md",
        gpgcheck=True,
        repo_gpgcheck=False,
    )
    rq = default_backend.Repoquery(bm.fill_sack())
    query = rq.query()
    assert len(query) == 2
    assert {p.name for p in query} == {"copr-module-integration-dummy-package"}



# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later


from pathlib import Path

from fedrq import config as rqconfig
from fedrq.backends.base import PackageCompat, PackageQueryCompat, RepoqueryBase


def test_make_base_rawhide_repos() -> None:
    config = rqconfig.get_config()
    rawhide = config.get_release("rawhide")
    bm = config.backend_mod.BaseMaker()
    base = rawhide.make_base(config, fill_sack=False, base_maker=bm)  # noqa: F841
    repos = bm.repolist(True)
    assert len(repos) == len(rawhide.repog.repos)
    assert set(repos) == set(rawhide.repog.repos)


def test_package_protocol(repo_test_rq: RepoqueryBase):
    package = repo_test_rq.get_package("packagea", arch="noarch")
    assert isinstance(package, PackageCompat)


def test_query_protocol(repo_test_rq: RepoqueryBase):
    query = repo_test_rq.query()
    assert isinstance(query, PackageQueryCompat)


def test_baseurl_repog(repo_test_rq: RepoqueryBase, data_path: Path):
    for i in ("", "file://"):
        second_rq = rqconfig.get_config().get_rq(
            "rawhide", f"@baseurl:{i}{data_path/ 'repos' / 'repo1' / 'repo'}"
        )
        assert sorted(map(str, repo_test_rq.query())) == sorted(
            map(str, second_rq.query())
        )



# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later

"""
This module houses code to load configuration from the filesystem and validate
it.
"""

from __future__ import annotations

import importlib.resources as importlib_resources
import itertools
import logging
import os
import re
import sys
import typing as t
import warnings
import zipfile
from collections.abc import Callable
from enum import auto as auto_enum
from importlib.abc import Traversable
from pathlib import Path

if sys.version_info < (3, 11):
    import tomli as tomllib
else:
    import tomllib

from pydantic import BaseModel, Field, PrivateAttr, validator

from fedrq._compat import StrEnum
from fedrq._config import ConfigError
from fedrq._utils import merge_dict, mklog
from fedrq.backends import BACKENDS, get_default_backend
from fedrq.backends.base import BaseMakerBase
from fedrq.release_repo import AliasRepoG, DefaultRepoGs, RepoG, Repos

if t.TYPE_CHECKING:
    import dnf
    import libdnf5

    from fedrq.backends.base import BackendMod, RepoqueryBase

CONFIG_DIRS = (Path.home() / ".config/fedrq", Path("/etc/fedrq"))
DEFAULT_REPO_CLASS = "base"
DEFAULT_COPR_BASEURL = "https://copr.fedoraproject.org"
logger = logging.getLogger(__name__)


class LoadFilelists(StrEnum):
    auto = auto_enum()
    always = auto_enum()
    never = auto_enum()

    @classmethod
    def from_bool(cls, /, boolean: bool) -> LoadFilelists:
        return cls.always if boolean else cls.never

    def __bool__(self) -> bool:
        return self == LoadFilelists.always


class ReleaseConfig(BaseModel):
    name: str = Field(exclude=True)
    defs: dict[str, list[str]]
    version: t.Optional[str] = None
    matcher: t.Pattern
    repo_dirs: list[Path] = Field(
        default_factory=lambda: [
            directory.joinpath("repos") for directory in CONFIG_DIRS
        ]
    )
    defpaths: set[str] = Field(default_factory=set)
    system_repos: bool = True
    append_system_repos: bool = False

    koschei_collection: t.Optional[str] = None
    copr_chroot_fmt: t.Optional[str] = None

    full_def_paths: t.ClassVar[list[t.Union[Traversable, Path]]] = []
    repo_aliases: dict[str, str] = {}
    _repogs = PrivateAttr()

    @property
    def repogs(self) -> Repos:
        return self._repogs

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._repogs = DefaultRepoGs.new(
            self.defs | AliasRepoG.from_str_mapping(self.repo_aliases)
        )

    @validator("defpaths")
    def v_defpaths(cls, value, values) -> dict[str, t.Any]:
        flog = mklog(__name__, "ReleaseConfig", "_get_full_defpaths")
        flog.debug(f"Getting defpaths for {values['name']}: {value}")
        values["full_def_paths"] = cls._get_full_defpaths(
            values["name"], value.copy(), values["repo_dirs"]
        )
        return value

    @validator("matcher")
    def v_matcher(cls, value: t.Pattern, values: dict[str, t.Any]) -> t.Pattern:
        if not values["version"] and value.groups != 1:
            raise ValueError("'matcher' must have exactly one capture group")
        return value

    @validator("repo_dirs", pre=True)
    def v_repo_dirs(cls, value: str | list[Path]) -> list[Path]:
        if not isinstance(value, str):
            return value
        return [Path(directory) for directory in value.split(":")]

    @validator("append_system_repos", always=True)
    def v_append_system_repos(cls, value: bool, values: dict[str, t.Any]) -> bool:
        if value:
            values["system_repos"] = True
        return value

    def is_match(self, val: str) -> bool:
        return bool(re.fullmatch(self.matcher, val))

    def is_valid_repo(self, val: str) -> bool:
        try:
            self.repogs.get_repo(val)
        except ConfigError:
            return False
        else:
            return True

    @staticmethod
    def _repo_dir_iterator(
        repo_dirs: list[Path],
    ) -> t.Iterator[t.Union[Traversable, Path]]:
        flog = mklog(__name__, "ReleaseConfig", "_repo_dir_iterator")
        topdirs: tuple[t.Union[Traversable, Path], ...] = (
            *repo_dirs,
            importlib_resources.files("fedrq.data.repos"),
        )
        flog.debug("topdirs = %s", topdirs)
        for topdir in topdirs:
            if isinstance(topdir, Path):
                topdir = topdir.expanduser()
            if not topdir.is_dir():
                continue
            for file in topdir.iterdir():
                if file.is_file():
                    yield file

    @classmethod
    def _get_full_defpaths(
        cls, name: str, defpaths: set[str], repo_dirs: list[Path]
    ) -> list[t.Union[Traversable, Path]]:
        missing_absolute: list[t.Union[Traversable, Path]] = []
        full_defpaths: list[t.Union[Traversable, Path]] = []
        flog = mklog(__name__, cls.__name__, "_get_full_defpaths")
        flog.debug(f"Searching for absolute defpaths: {defpaths}")
        for defpath in defpaths.copy():
            if (path := Path(defpath).expanduser()).is_absolute():
                flog.debug(f"Is absolute: {path}")
                defpaths.discard(defpath)
                if path.is_file():
                    flog.debug(f"Exists: {path}")
                    full_defpaths.append(path)
                else:
                    flog.debug(f"Doesn't Exist: {path}")
                    missing_absolute.append(path)
        flog.debug(f"Getting relative defpaths: {defpaths}")
        files = cls._repo_dir_iterator(repo_dirs)
        while defpaths:
            try:
                file = next(files)
                flog.debug(f"file={file}")
            except StopIteration:
                flog.debug(msg="StopIteration")
                break
            if file.name in defpaths:
                flog.debug(f"{file.name} in {defpaths}")
                full_defpaths.append(file)
                defpaths.discard(file.name)
        if defpaths:
            _missing = ", ".join(
                sorted(str(p) for p in ((*defpaths, *missing_absolute)))
            )
            raise ConfigError(f"Missing defpaths in {name}: {_missing}")
        return full_defpaths

    def get_release(
        self, config: RQConfig, branch: str, repo_name: str = "base"
    ) -> Release:
        return Release(
            config=config, release_config=self, branch=branch, repo_name=repo_name
        )


class Release:
    """
    Encapsulates a ReleaseConfig with a specific version and repo name.
    This SHOULD NOT be instantiated directly.
    The __init__() has no stability promises.
    Use the [`RQConfig.get_release()`][fedrq.config.RQConfig.get_release]
    factory instead.
    """

    def __init__(
        self,
        config: RQConfig,
        release_config: ReleaseConfig,
        branch: str,
        repo_name: str = "base",
    ) -> None:
        self.config = config
        self.release_config = release_config
        if not self.release_config.is_match(branch):
            raise ConfigError(
                f"Branch {branch} does not match {self.release_config.name}"
            )
        self.branch = branch
        self.repo_name = repo_name
        self.repog: RepoG = self.get_repog(repo_name)

    def get_repog(self, key: str) -> RepoG:
        return self.release_config.repogs.get_repo(key)

    @property
    def version(self) -> str:
        v: str | None = None
        if self.release_config.version:
            v = self.release_config.version
        elif match := re.fullmatch(self.release_config.matcher, self.branch):
            v = match.group(1)
        if v is None:
            raise ValueError(f"{self.branch} does not match {self.release_config.name}")
        # Special case
        if v == "$releasever":
            v = self.config.backend_mod.get_releasever()
        return v

    @property
    def copr_chroot_fmt(self) -> str | None:
        return self.release_config.copr_chroot_fmt

    @property
    def koschei_collection(self) -> str | None:
        return self.release_config.koschei_collection

    def make_base(
        self,
        config: RQConfig | None = None,
        base_conf: dict[str, t.Any] | None = None,
        base_vars: dict[str, t.Any] | None = None,
        base_maker: BaseMakerBase | None = None,
        fill_sack: bool = True,
    ) -> dnf.Base | libdnf5.base.Base:
        """
        :param config: An RQConfig object. Not passing this argument is deprecated.
                       DEPRECATED since 0.4.0: A new RQConfig object will be
                                               created if this is None.
        :param base_conf: Base session configuration
        :param base_vars: Base session vars/substitutions (arch, basearch,
                                                           releasever, etc.)
        :param base_maker: Existing BaseMaker object to configure.
                           If base_maker is None, a new one will be created.
        :param fill_sack: Whether to fill the Base object's package sack or
                          just return the Base object after applying configuration.
        """
        if config is None:
            warnings.warn(
                "DEPRECATED since 0.4.0: Provide an RQConfig object for 'config'"
            )
            config = get_config()
        base_conf = base_conf or {}
        base_vars = base_vars or {}
        releasever = config.backend_mod.get_releasever()
        if (
            "cachedir" not in base_conf
            and config.smartcache
            and self.version != releasever
        ):
            logger.debug("Using smartcache")
            base_conf["cachedir"] = str(get_smartcache_basedir() / str(self.version))
        bm = base_maker or config.backend_mod.BaseMaker()
        bm.sets(base_conf, base_vars)
        if config.load_filelists:
            bm.load_filelists()
        bm.load_release_repos(self, "releasever" not in base_vars)
        return bm.fill_sack() if fill_sack else bm.base

    def _copr_repo(
        self, value: str, default_copr_baseurl: str = DEFAULT_COPR_BASEURL
    ) -> str:
        value = value.rstrip("/")
        if not self.copr_chroot_fmt:
            raise ValueError(
                f"{self.release_config.name} does not have 'copr_chroot_fmt' set"
            )
        chroot = re.sub("-{arch}$", "", self.copr_chroot_fmt).format(
            version=self.version
        )
        if value.startswith(("http://", "https://")):
            return value + "/" + chroot

        frag = "coprs/"
        if value.startswith("@"):
            frag += "g/"
            value = value[1:]
        value, sep, copr_baseurl = value.partition("@")
        if not sep:
            copr_baseurl = default_copr_baseurl.rstrip("/")
        elif not copr_baseurl.startswith(("http://", "https://")):
            copr_baseurl = "https://" + copr_baseurl
        frag += value
        return f"{copr_baseurl}/{frag}/repo/{chroot}"


class RQConfig(BaseModel):
    backend: t.Optional[str] = os.environ.get("FEDRQ_BACKEND")
    releases: dict[str, ReleaseConfig]
    default_branch: str = os.environ.get("FEDRQ_BRANCH", "rawhide")
    smartcache: bool = True
    load_filelists: LoadFilelists = LoadFilelists.auto
    _backend_mod = None
    copr_baseurl: str = DEFAULT_COPR_BASEURL

    class Config:
        json_encoders: dict[t.Any, Callable[[t.Any], str]] = {
            re.Pattern: lambda pattern: pattern.pattern,
            zipfile.Path: lambda path: str(path),
        }
        underscore_attrs_are_private = True
        validate_assignment = True

    @validator("backend")
    def v_backend(cls, value) -> str:
        assert (
            value is None or value in BACKENDS
        ), f"Valid backends are: {', '.join(BACKENDS)}"
        return value

    @property
    def backend_mod(self) -> BackendMod:
        if not self._backend_mod:
            self._backend_mod = get_default_backend(
                self.backend,
                # allow falling back to a non default backend
                # (i.e. not backends.DEFAULT_BACKEND)
                # when the user does not explicitly request a backend.
                fallback=not bool(self.backend),
            )
        return self._backend_mod

    def get_release(
        self, branch: str | None = None, repo_name: str | None = None
    ) -> Release:
        flog = mklog(__name__, "RQConfig", "get_releases")
        branch = branch or self.default_branch
        repo_name = repo_name or DEFAULT_REPO_CLASS
        pair = (branch, repo_name)
        for release in sorted(
            self.releases.values(), key=lambda r: r.name, reverse=True
        ):
            try:
                r = release.get_release(self, branch=branch, repo_name=repo_name)
            except ConfigError as exc:
                logger.debug(f"{release.name} does not match {pair}: {exc}")
            else:
                flog.debug("%s matches %s", release.name, pair)
                return r
        raise ConfigError(
            "{} does not much any of the configured releases: {}".format(
                pair, self.release_names
            )
        )

    @property
    def release_names(self) -> list[str]:
        return [rc.name for rc in self.releases.values()]

    def get_rq(
        self,
        branch: str | None = None,
        repo: str | None = None,
        base_conf: dict[str, t.Any] | None = None,
        base_vars: dict[str, t.Any] | None = None,
    ) -> RepoqueryBase:
        """
        Higher level interface that finds the Release object that mathces
        {branch} and {repo}, creates a (lib)dnf(5).base.Base session, and
        returns a Repoquery object.

        :param branch: branch name
        :param repo: repo class. defaults to 'base'.
        :param base_conf: Base session configuration
        :param base_vars: Base session vars/substitutions (arch, basearch,
                                                           releasever, etc.)
        """
        release = self.get_release(branch, repo)
        return self.backend_mod.Repoquery(release.make_base(self, base_conf, base_vars))


def get_smartcache_basedir() -> Path:
    basedir = Path(os.environ.get("XDG_CACHE_HOME", Path("~/.cache").expanduser()))
    return basedir.joinpath("fedrq").resolve()


def _get_files(
    directory: t.Union[Traversable, Path], suffix: str, reverse: bool = True
) -> list[t.Union[Traversable, Path]]:
    files: list[t.Union[Traversable, Path]] = []
    if not directory.is_dir():
        return files
    for file in directory.iterdir():
        if file.name.endswith(suffix) and file.is_file():
            files.append(file)
    return sorted(files, key=lambda f: f.name, reverse=reverse)


def get_config(**overrides: t.Any) -> RQConfig:
    """
    Retrieve config files from CONFIG_DIRS and fedrq.data.
    Perform naive top-level merging of the 'releases' table.
    """
    flog = mklog(__name__, "get_config")
    flog.debug(f"CONFIG_DIRS = {CONFIG_DIRS}")
    config: dict[str, t.Any] = {}
    all_files: list[list[t.Union[Traversable, Path]]] = [
        _get_files(importlib_resources.files("fedrq.data"), ".toml"),
        *(_get_files(p, ".toml") for p in reversed(CONFIG_DIRS)),
    ]
    flog.debug("all_files = %s", all_files)
    for path in itertools.chain.from_iterable(all_files):
        flog.debug("Loading config file: %s", path)
        with path.open("rb") as fp:
            data = tomllib.load(fp)
        merge_dict(data, config)
    merge_dict(overrides, config)
    config["releases"] = _get_releases(config["releases"])
    flog.debug("Final config: %s", config)
    return RQConfig(**config)


def _get_releases(rdict: dict[str, dict[str, t.Any]]) -> dict[str, t.Any]:
    releases: dict[str, t.Any] = {}
    for name, data in rdict.items():
        releases[name] = dict(name=name, **data)
    return releases


def get_rq(
    branch: str | None = None,
    repo: str = "base",
    *,
    smart_cache: bool | None = None,
    load_filelists: bool | None = None,
) -> RepoqueryBase:
    """
    **DEPRECATED since 0.4.0.** Use
    [`RQConfig.get_rq()`][fedrq.config.RQConfig.get_rq] instead.

    Higher level interface that creates an RQConfig object, finds the Release
    object that mathces {branch} and {repo}, creates a dnf.Base, and finally
    returns a Repoquery object.
    """
    warnings.warn("DEPRECATED since 0.4.0: use RQConfig.get_rq() instead.")
    config = get_config()
    if smart_cache is not None:
        config.smartcache = smart_cache
    if load_filelists is not None:
        config.load_filelists = LoadFilelists.from_bool(load_filelists)
    return config.get_rq(branch, repo)



# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later

import glob

TEST_REPO = "repo1"


def test_repo_package_count(repo_test_rq, data_path):
    directory = data_path / "repos" / "repo1"
    count_specs = len(glob.glob(str(directory / "specs/**/*.spec"), recursive=True))
    assert count_specs == len(repo_test_rq.query(arch="src"))
    assert count_specs == 3



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from collections.abc import Callable, Sequence

import pytest

import fedrq.backends
import fedrq.cli
from fedrq.backends.base import BackendMod


def test_disablerepo_wildcard(patch_config_dirs, capsys):
    fedrq.cli.main(["pkgs", "--disablerepo=*", "*"])
    out, err = capsys.readouterr()
    assert not out and not err


@pytest.mark.parametrize(
    "args",
    [
        pytest.param(["-r", "testrepo1"]),
        pytest.param(["--disablerepo=*", "-e", "testrepo1"]),
        pytest.param(["--disablerepo=*", "--enablerepo", "base"]),
        pytest.param(["--disablerepo=*", "--enablerepo", "@base"]),
    ],
)
def test_repo_by_real_name(patch_config_dirs, capsys, args):
    fedrq.cli.main(["pkgs", *args, "-F", "repoid", "*"])
    out, err = capsys.readouterr()
    assert set(out.splitlines()) == {"testrepo1"}
    assert not err


@pytest.mark.no_rpm_mock
@pytest.mark.parametrize(
    ["args", "releasever"],
    [
        pytest.param(
            ["-b", "local", "ansible"], lambda backend: backend.get_releasever()
        ),
        pytest.param(["-b", "local:34", "ansible"], lambda _: "34"),
    ],
)
def test_local_release(
    monkeypatch: pytest.MonkeyPatch,
    args: Sequence[str],
    releasever: Callable[[BackendMod], str],
):
    backend = fedrq.backends.get_default_backend()
    base_maker = backend.BaseMaker()
    base_maker.read_system_repos(False)
    repolist = base_maker.repolist()
    del base_maker

    parser = fedrq.cli.Pkgs.make_parser()
    namespace = parser.parse_args(args)
    # Calling fill_sack() is expensive and not necessary here
    monkeypatch.setattr(backend.BaseMaker, "fill_sack", lambda self: self.base)
    obj = fedrq.cli.Pkgs(namespace)
    base_maker = backend.BaseMaker(obj.rq.base)
    assert sorted(base_maker.repolist()) == sorted(repolist)
    # TODO: Add a proper API for this
    if backend.BACKEND == "dnf":
        assert releasever(backend) == base_maker.base.conf.substitutions["releasever"]
    else:
        assert releasever(backend) == base_maker.base.get_vars().get_value("releasever")



# PYTHON_ARGCOMPLETE_OK
# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import abc
import argparse
import collections.abc as cabc
import json
import logging
import re
import sys
from functools import wraps
from pathlib import Path
from typing import TYPE_CHECKING, Any

from fedrq.backends.base import BaseMakerBase

try:
    import tomli_w
except ImportError:
    HAS_TOMLI_W = False
else:
    HAS_TOMLI_W = True

from pydantic import ValidationError

from fedrq._utils import mklog
from fedrq.backends import BACKENDS, MissingBackendError
from fedrq.cli.formatters import (
    DefaultFormatters,
    Formatter,
    FormatterError,
    Formatters,
)
from fedrq.config import ConfigError, LoadFilelists, Release, RQConfig, get_config

if TYPE_CHECKING:
    from fedrq.backends.base import BackendMod, PackageQueryCompat

logger = logging.getLogger("fedrq")

FORMATTER_ERROR_SUFFIX = "See fedrq(1) for more information about formatters."

MISSING_BACKEND_MSG = """
Failed to load the package management backend: {}
These modules are only available for the default system Python interpreter.
""".strip()


SPLIT_REGEX = re.compile(r"\s*[,\s]\s*")


# Based on dnf.cli.option_parser._RepoCallback
class _EnableDisableRepo(argparse.Action):
    OPERATORS = {"-e": "enable", "--enablerepo": "enable", "--disablerepo": "disable"}

    def __call__(
        self,
        parser: argparse.ArgumentParser,  # noqa: ARG002
        namespace: argparse.Namespace,
        values: str | cabc.Sequence[Any] | None,
        option_string: str | None = None,
    ) -> None:
        assert isinstance(values, str) and isinstance(option_string, str)
        operator = self.OPERATORS[option_string]
        getattr(namespace, self.dest).extend(
            (operator, repo) for repo in SPLIT_REGEX.split(values)
        )


def _append_error(lst: list[str], error: cabc.Iterable | str | None = None) -> None:
    if isinstance(error, str):
        lst.append(error)
    elif isinstance(error, cabc.Iterable):
        lst.append(*error)
    elif error:
        raise TypeError(f"{type(error)} is not a valid return type.")


def v_add_errors(func: cabc.Callable[..., str | cabc.Iterable | None]) -> cabc.Callable:
    @wraps(func)
    def wrapper(self: Command, *args, **kwargs) -> str | cabc.Iterable | None:
        error = func(self, *args, **kwargs)
        _append_error(self._v_errors, error)
        return error

    return wrapper


def v_fatal_error(
    func: cabc.Callable[..., str | cabc.Iterable | None]
) -> cabc.Callable:
    def wrapper(self: Command, *args, **kwargs) -> None:
        error = func(self, *args, **kwargs)
        fatal: list[str] = []
        _append_error(fatal, error)
        if not fatal:
            return None
        self._v_handle_errors(False)
        for err in fatal:
            print("FATAL ERROR:", err, file=sys.stderr)
        sys.exit(1)

    return wrapper


class Command(abc.ABC):
    config: RQConfig
    release: Release
    query: PackageQueryCompat
    formatters: Formatters = DefaultFormatters
    formatter: Formatter

    def __init__(self, args: argparse.Namespace):
        self.args = args

        self.v_logging()
        flog = mklog(__name__, self.__class__.__name__)
        flog.debug("args=%s", args)

        self.get_names()

        try:
            self.config = self._get_config()
        except ValidationError as exc:
            sys.exit(str(exc))
        self._set_config("backend")
        self._set_config("smartcache")
        if (
            self.args.load_filelists == LoadFilelists.auto
            and "files" in self.args.formatter
        ):
            self.config.load_filelists = LoadFilelists.always
        else:
            self._set_config("load_filelists")

        self._v_errors: list[str] = []

    def _get_config(self):
        # This makes it easier to mock the config
        return get_config()

    @property
    def backend(self) -> BackendMod:
        return self.config.backend_mod

    @abc.abstractmethod
    def run(self) -> None:
        ...

    def _set_config(self, key: str) -> None:
        arg = getattr(self.args, key, None)
        if arg is not None:
            setattr(self.config, key, arg)

    def _logq(
        self, query: PackageQueryCompat, name: str = "query", level=logging.DEBUG
    ) -> PackageQueryCompat:
        """
        Log a query object after performing the appropriate formatting.
        Don't iterate over the query unless the logging level is low enough to
        avoid a performance penalty.
        """
        if logger.getEffectiveLevel() <= level:
            logger.debug("%s = %s", name, tuple(query))
        return query

    @classmethod
    def branch_repo_parser(cls) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(add_help=False)
        parser.add_argument(
            "-b",
            "--branch",
            help="Fedora or EPEL branch name "
            "(e.g. epel7, rawhide, epel9-next, f37) to query",
        )
        parser.add_argument("-r", "--repos", default="base")
        parser.add_argument(
            "-e",
            "--enablerepo",
            dest="enable_disable",
            default=[],
            action=_EnableDisableRepo,
            metavar="REPO",
            help="""
            Enable certain repositories for the duration of this operation.
            All repositories in the system configuration and any additional
            defs in the selected branch are available.
            """,
        )
        parser.add_argument(
            "--disablerepo",
            dest="enable_disable",
            default=[],
            action=_EnableDisableRepo,
            metavar="REPO",
            # PROVISIONAL
            help=argparse.SUPPRESS,
        )
        return parser

    @classmethod
    def parent_parser(cls) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(
            add_help=False, parents=[cls.branch_repo_parser()]
        )
        parser.add_argument(  # type: ignore[attr-defined]
            "names", metavar="NAME", nargs="*", help="Mutually exclusive with --stdin"
        ).completer = lambda **_: ()
        parser.add_argument(
            "-i", "--stdin", help="Read package names from stdin.", action="store_true"
        )
        parser.add_argument("-l", "--latest", default=1, help="'all' or an intenger")
        parser.add_argument(  # type: ignore[attr-defined]
            "-F",
            "--formatter",
            default="plain",
        ).completer = cls.formatters._argcompleter
        cachedir_group = parser.add_mutually_exclusive_group()
        cachedir_group.add_argument(
            "--system-cache",
            action="store_true",
            help="Use the default dnf cachedir and ignore `smartcache` config option",
        )
        cachedir_group.add_argument(
            "--sc",
            "--smartcache",
            action="store_true",
            default=None,
            dest="smartcache",
            help="See `smartcache` in fedrq(5)."
            " smartcache is enabled by default,"
            " so this is noop unless you set `smartcache=false` in the config file.",
        )
        # This is mutually exclusive with --smartcache. It's still undocumented
        # and subject to change.
        cachedir_group.add_argument("--cachedir", help=argparse.SUPPRESS, type=Path)
        parser.add_argument("--debug", action="store_true")
        parser.add_argument(
            "-L",
            "--filelists",
            choices=tuple(LoadFilelists),
            dest="load_filelists",
            help="Whether to load filelists."
            " By default, filelists are only loaded when using the files formatter."
            " This only applies when using the libdnf5 backend,"
            " which doesn't load filelists by default to save memory and bandwidth."
            " dnf4 always loads filelists.",
        )
        parser.add_argument("-B", "--backend", choices=tuple(BACKENDS))
        parser.add_argument(
            "--forcearch", help="Query a foreign architecture's repositories"
        )
        return parser

    @classmethod
    @abc.abstractmethod
    def make_parser(
        cls,
        parser_func: cabc.Callable = argparse.ArgumentParser,
        *,
        add_help: bool = False,
        **kwargs,
    ) -> argparse.ArgumentParser:
        ...

    @classmethod
    def standalone(cls, argv: list[str] | None = None) -> None:
        parser = cls.make_parser(add_help=False)
        return cls(args=parser.parse_args(argv)).run()

    def get_names(self) -> None:
        if self.args.names and self.args.stdin:
            sys.exit("Postional NAMEs can not be used with --stdin")
        if self.args.stdin:
            self.args.names = [line.strip() for line in sys.stdin.readlines()]
        if not self.args.names:
            sys.exit("No package names were passed")

    def format(self) -> cabc.Iterable[str]:
        """
        Helper to run `self.formatter.format(self.query)`
        """
        return self.formatter.format(self.query)

    def _v_handle_errors(self, should_exit: bool = True):
        if self._v_errors:
            for line in self._v_errors:
                print("ERROR:", line, file=sys.stderr)
            if should_exit:
                sys.exit(1)

    def v_logging(self) -> None:
        if getattr(self.args, "debug", None):
            logger.setLevel(logging.DEBUG)

    @v_add_errors
    def v_latest(self) -> str | None:
        try:
            self.args.latest = int(self.args.latest)
        except ValueError:
            if isinstance(self.args.latest, str) and self.args.latest.lower() in (
                "a",
                "all",
            ):
                self.args.latest = None
            else:
                return "--latest must equal 'all' or be an integer"
        return None

    @v_add_errors
    def v_formatters(self) -> str | None:
        try:
            self.formatter = self.formatters.get_formatter(self.args.formatter)
        except FormatterError as err:
            return str(err) + "\n" + FORMATTER_ERROR_SUFFIX
        return None

    @v_add_errors
    def v_arch(self) -> str | None:
        # TODO: Verify that arches are actually valid RPM arches.
        if not self.args.arch:
            return None
        if "notsrc" in self.args.arch and "," in self.args.arch:
            return (
                f"Illegal option '--arch={self.args.arch}': "
                "'notsrc' is a special keyword that cannot be part of a list"
            )
        if "," in self.args.arch:
            self.args.arch = [item.strip() for item in self.args.arch.split(",")]
        return None

    @v_fatal_error
    def v_release(self) -> str | None:
        try:
            self.release = self.config.get_release(self.args.branch, self.args.repos)
        except ConfigError as err:
            return str(err)
        return None

    def _enable_disable_bm(self, bm: BaseMakerBase):
        for func, repo in self.args.enable_disable:
            if func == "enable":
                self.release.get_repog(repo).load(bm, self.config, self.release)
            elif func == "disable":
                bm.disable_repo(repo, True)
            else:
                raise ValueError

    @v_fatal_error
    def v_rq(self) -> str | None:
        conf: dict[str, Any] = {}
        bvars: dict[str, Any] = {}

        # Set cachedir if it's explicitly passed
        if self.args.cachedir:
            conf["cachedir"] = str(self.args.cachedir)
        # Disable release based smartcache if user explicitly disabled it or if
        # forcearch is in use.
        elif self.args.system_cache or self.args.forcearch:
            self.config.smartcache = False

        if self.args.forcearch:
            conf["ignorearch"] = True
            bvars["arch"] = self.args.forcearch
        bm = self.backend.BaseMaker()
        try:
            self.release.make_base(self.config, conf, bvars, bm, False)
            self._enable_disable_bm(bm)
        except ConfigError as exc:
            return str(exc)
        self.rq = self.backend.Repoquery(bm.fill_sack())
        return None

    @v_fatal_error
    def v_backend(self) -> str | None:
        try:
            _ = self.backend
        except MissingBackendError as exc:
            return MISSING_BACKEND_MSG.format(str(exc))
        return None

    def v_default(self) -> None:
        self.v_formatters()
        self.v_latest()
        self.v_arch()
        # Fatal
        self.v_backend()
        self.v_release()
        self.v_rq()
        self._v_handle_errors()


class CheckConfig(Command):
    """
    Verify fedrq configuration
    """

    def __init__(self, args: argparse.Namespace):
        self.args = args
        self.v_logging()

    @staticmethod
    def _strip_nones(dct: dict[Any, Any], level=0) -> dict[Any, Any]:
        """
        Recurisvely remove None values from a dictionary so they can be TOML
        serialized
        """
        flog = mklog("fedrq.cli.CheckConfig", "_strip_nones_")
        for k in tuple(dct):
            if dct[k] is None:
                flog.debug("%s: Strip %s key", level, k)
                del dct[k]
            elif isinstance(dct[k], dict):
                flog.debug("%s: Recursing through %s dict", level, k)
                CheckConfig._strip_nones(dct[k], level + 1)
        return dct

    @classmethod
    def make_parser(
        cls,
        parser_func: cabc.Callable = argparse.ArgumentParser,
        *,
        add_help: bool = False,
        **kwargs,
    ) -> argparse.ArgumentParser:
        kwargs = dict(description=cls.__doc__, **kwargs)
        if add_help:
            kwargs["help"] = cls.__doc__
        parser = parser_func(**kwargs)
        parser.add_argument("--debug", action="store_true")
        parser.add_argument(
            "--dump",
            action="store_true",
            help="Dump config as a toml file. Requires tomli-w.",
        )
        return parser

    def run(self):
        flog = mklog("fedrq.cli.CheckConfig")
        if self.args.dump and not HAS_TOMLI_W:
            sys.exit("tomli-w is required for --dump.")
        if not self.args.dump:
            print("Validating config...")
        try:
            self.config = get_config()
        except ValidationError as exc:
            sys.exit(str(exc))
        try:
            self.config.get_release(self.config.default_branch)
        except ConfigError:
            sys.exit(f"default_branch '{self.config.default_branch}' is invalid")
        if not self.args.dump:
            print("No validation errors found!")
        else:
            flog.debug("Removing Nones from configuration dict")
            data_dict = self._strip_nones(json.loads(self.config.json()))
            tomli_w.dump(data_dict, sys.stdout.buffer)


# class CommandBoilerplate(Command):
#     """
#     Help text
#     """
#     def __init__(self, args: argparse.Namespace) -> None:
#         super().__init__(args)
#         self.v_default()

#     @classmethod
#     def make_parser(
#         cls,
#         parser_func: cabc.Callable = argparse.ArgumentParser,
#         *,
#         add_help: bool = False,
#         **kwargs,
#     ) -> argparse.ArgumentParser:
#         kwargs["description"] = cls.__doc__
#         kwargs["parents"] = [cls.parent_parser()]
#         if add_help:
#             kwargs["help"] = cls.__doc__

#         parser = parser_func(**kwargs)
#         ...
#         return parser

#     def run(self) -> None:
#         ...



# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later

import pytest

import fedrq.cli


@pytest.mark.skip(
    "This test loads rawhide metadata and is expensive. We already load f37."
)
@pytest.mark.no_rpm_mock
def test_pkgs_basic_rawhide(capsys, target_cpu):
    fedrq.cli.main(["pkgs", "bash", "-Fna", "--sc"])
    stdout, stderr = capsys.readouterr()
    assert sorted(stdout.splitlines()) == sorted(["bash.src", f"bash.{target_cpu}"])


@pytest.mark.no_rpm_mock
def test_pkgs_forcearch(runs):
    stdout, stderr = runs(
        ["pkgs", "--forcearch", "s390x", "-F", "arch", "-b", "f37", "*"], False
    )
    assert not stderr
    assert set(stdout) == {"noarch", "s390x", "src"}



# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import argparse
from collections.abc import Callable

from fedrq.cli.base import Command, v_add_errors


class Subpkgs(Command):
    """
    For each SRPM name, list the subpackages that it provides
    """

    def __init__(self, args: argparse.Namespace) -> None:
        super().__init__(args)
        self.v_default()

    @classmethod
    def make_parser(
        cls,
        parser_func: Callable = argparse.ArgumentParser,
        *,
        add_help: bool = False,
        **kwargs,
    ) -> argparse.ArgumentParser:
        pargs = dict(description=cls.__doc__, parents=[cls.parent_parser()], **kwargs)
        if add_help:
            pargs["help"] = "Find the subpackages of a list of SRPMs"
        parser = parser_func(**pargs)
        parser.add_argument(
            "-M",
            "--match",
            action="append",
            help="Only show subpackages whose name matches this string."
            " Glob patterns are permitted."
            " When specified multiple times, _any_ match is included.",
        )
        arch_group = parser.add_mutually_exclusive_group()
        arch_group.add_argument(
            "-A", "--arch", help="Only show subpackages with this arch"
        )
        arch_group.add_argument(
            "-S",
            "--notsrc",
            dest="arch",
            action="store_const",
            const="notsrc",
            help="This includes all binary RPMs. Multilib is excluded on x86_64. "
            "Equivalent to --arch=notsrc",
        )
        return parser

    def run(self) -> None:
        srpms = self.rq.resolve_pkg_specs(self.args.names, latest=self.args.latest)
        srpms.filterm(arch="src")
        self.query = self.rq.get_subpackages(
            srpms, latest=self.args.latest, arch=self.args.arch
        )
        if self.args.match:
            self.query.filterm(name__glob=self.args.match)
        for p in self.format():
            print(p)

    @v_add_errors
    def v_arch(self) -> str | None:
        if super().v_arch():
            return None
        if (
            self.args.arch
            and "src" in self.args.arch
            and "notsrc" not in self.args.arch
        ):
            return "Illegal option '--arch=src': Subpackages are binary RPMs"
        return None



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-2.0-or-later

import pytest

import fedrq.cli


def test_repolist(capsys: pytest.CaptureFixture):
    fedrq.cli.main(
        [
            "repolist",
            "--branch=rawhide",
            "--repos=buildroot",
            "--disablerepo=fedrq-koji-rawhide-source",
            "--enablerepo=@repo:fedrq-koji-rawhide-source",
        ]
    )
    out, err = capsys.readouterr()
    assert sorted(out.splitlines()) == [
        "fedrq-koji-rawhide",
        "fedrq-koji-rawhide-source",
    ]
    assert not err.strip()



# Copyright (C) 2023 Maxwell G <gotmax@e.email>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ruff: noqa: ARG002

from __future__ import annotations

import abc
import logging
import os
import tempfile
from collections.abc import (
    Callable,
    ItemsView,
    Iterable,
    Iterator,
    KeysView,
    Mapping,
    Sequence,
    ValuesView,
)
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, NoReturn

import requests

from fedrq._config import ConfigError

if TYPE_CHECKING:
    from fedrq.backends.base import BaseMakerBase
    from fedrq.config import Release, RQConfig

LOG = logging.getLogger(__name__)


@contextmanager
def _get_file(path: str) -> Iterator[str]:
    if path.startswith(("http://", "https://")):
        try:
            name = None
            LOG.info("Downloading %s", path)
            req = requests.get(path)
            if req.status_code != 200:
                raise ConfigError(f"Failed to download {path}")
            fd, name = tempfile.mkstemp()
            os.write(fd, req.content)
            os.close(fd)
            yield name
        finally:
            if name:
                os.unlink(name)
    elif path.startswith("file://"):
        yield path[7:]
    else:
        yield path


class RepoG(metaclass=abc.ABCMeta):
    """
    Base class containing a repo group to load. These can be added to a Repos
    container class.
    """

    name: str
    seperator: str
    args: str
    config: RQConfig
    release: Release
    container: Repos

    def __init__(
        self,
        name: str,
        seperator: str,
        args: str,
        container: Repos | None = None,
    ) -> None:
        self.name = name
        self.seperator = seperator
        self.args = args
        self.container = container or Repos({})
        self.validate()

    @abc.abstractmethod
    def load(
        self, base_maker: BaseMakerBase, config: RQConfig, release: Release
    ) -> None:
        ...

    def err_fmt(self, msg: str) -> ConfigError:
        return ConfigError(msg)

    def err(self, msg: str, from_value: Any = ...) -> NoReturn:
        if from_value == ...:
            raise self.err_fmt(msg)
        else:
            raise self.err_fmt(msg) from from_value

    def validate(self) -> None:
        if not self.seperator and not self.args:
            self.err("Expected an argument")

    def __str__(self) -> str:
        return f"{self.name}{self.seperator}{self.args}"


class SimpleRepoG(RepoG):
    def load(
        self, base_maker: BaseMakerBase, config: RQConfig, release: Release
    ) -> None:
        try:
            base_maker.enable_repo(self.args)
        except ValueError:
            self.err(f"No repo named {self.args}", None)


class MultiNameG(RepoG):
    repos: Sequence[str] = ()
    repogs: list[RepoG]

    def validate(self):
        if self.args:
            raise ConfigError("No arguments are accepted")
        self.repogs = [self.container.get_repo(repo) for repo in self.repos]

    def load(
        self, base_maker: BaseMakerBase, config: RQConfig, release: Release
    ) -> None:
        for repog in self.repogs:
            repog.load(base_maker, config, release)

    @classmethod
    def from_names(
        cls, class_name: str, names: Sequence[str] | str
    ) -> type[MultiNameG]:
        dct = dict(
            __doc__=f"Load the following repos: {names}",
            __module__=__name__,
            repos=[names] if isinstance(names, str) else names,
        )
        typ = type(f"{class_name.upper()}MultiNameG", (cls,), dct)
        return typ


class AliasRepoG(RepoG):
    fmt_str: str
    final: RepoG

    @classmethod
    def from_str(cls, fmt_str: str, class_name: str) -> type[AliasRepoG]:
        dct = dict(
            __doc__=f"Load a repo from the {fmt_str!r} alias",
            __module__=__name__,
            fmt_str=fmt_str,
        )
        typ = type(f"{class_name.upper()}AliasRepoG", (cls,), dct)
        return typ

    @classmethod
    def from_str_mapping(cls, mapping: dict[str, str]) -> dict[str, type[AliasRepoG]]:
        return {name: cls.from_str(fmt_str, name) for name, fmt_str in mapping.items()}

    def validate(self) -> None:
        super().validate()
        expanded = self.fmt_str.format(*self.args.split(";"))
        self.final = self.container.get_repo(expanded)

    def load(
        self, base_maker: BaseMakerBase, config: RQConfig, release: Release
    ) -> None:
        self.final.load(base_maker, config, release)


class FileRepoG(RepoG):
    def load(
        self, base_maker: BaseMakerBase, config: RQConfig, release: Release
    ) -> None:
        with _get_file(self.args) as path:
            base_maker.read_repofile(path)


class CoprRepoG(RepoG):
    url: str

    def load(
        self, base_maker: BaseMakerBase, config: RQConfig, release: Release
    ) -> None:
        self.url = release._copr_repo(self.args, config.copr_baseurl)
        with _get_file(self.url) as path:
            base_maker._read_repofile_new(path, True)


def _clean_invalid(string: str, valid_chars: Iterable[str], repl: str) -> str:
    final = ""
    for char in string:
        if char not in valid_chars:
            char = repl
        final += char
    return final


class BaseurlRepoG(RepoG):
    _ALLOWED_REPOID_CHARS = (
        "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_.:"
    )
    _ATTR = "baseurl"
    _COERCE_TO_LIST: bool = True

    def load(
        self, base_maker: BaseMakerBase, config: RQConfig, release: Release
    ) -> None:
        url, _, key = self.args.partition(";")
        repoid = _clean_invalid(url, self._ALLOWED_REPOID_CHARS, "_")
        repo_kwargs: dict[str, Any] = {
            self._ATTR: ([url] if self._COERCE_TO_LIST else url),
            "gpgcheck": bool(key),
            "name": repoid,
        }
        if key:
            repo_kwargs["gpgkey"] = key
        base_maker.create_repo(repoid, **repo_kwargs)


class MirrorlistRepoG(BaseurlRepoG):
    _ATTR = "mirrorlist"
    _COERCE_TO_LIST = False


class Repos(Mapping[str, type[RepoG]]):
    """
    Immutable mapping like class of RepoG types.
    Converts repo aliases (strings) and list of repos in RepoG objects.
    Allows merging and adding other Repos objects.
    """

    # Normally, get_repo() will only select RepoGs in the container if the key
    # starts with '@' to avoid conflicts with plain repoids.
    # However, get_repo() will return a RepoG that's a subclass of
    # _ALLOWED_PLAIN even if the key doesn't start with '@'.
    _ALLOWED_PLAIN: tuple[type[RepoG]] = (MultiNameG,)
    # This RepoG will be used when get_repo() is passed a key that doesn't
    # start with '@'.
    _DEFAULT: type[RepoG] = SimpleRepoG
    # Factory function to generate a RepoG from a plain string or list.
    _FALLBACK_FACTORY: Callable[
        [str, Sequence[str] | str], type[RepoG]
    ] = MultiNameG.from_names

    def __init__(
        self,
        repo_classes: Mapping[str, Sequence[str] | str | type[RepoG]],
    ) -> None:
        self.__data: dict[str, type[RepoG]] = {
            name: repos
            if isinstance(repos, type) and issubclass(repos, RepoG)
            else self._FALLBACK_FACTORY(name, repos)
            for name, repos in ItemsView(repo_classes)
        }

    def get_repo(self, key: str) -> RepoG:
        if key.startswith("@"):
            name, seperator, args = key.partition(":")
            try:
                typ = self[name[1:]]
            except KeyError:
                raise ConfigError(f"{key} is not a valid repository class") from None
            return typ(name, seperator, args, self)
        # Repo groups are special cased to maintain backwards compatibility
        elif key in self.keys() and issubclass(self[key], self._ALLOWED_PLAIN):
            return self[key](key, "", "", self)
        else:
            return self._DEFAULT("", "", key, self)

    def __getitem__(self, key: str) -> type[RepoG]:
        return self.__data[key]

    def __len__(self) -> int:
        return len(self.__data)

    def __iter__(self) -> Iterator[str]:
        return iter(self.__data)

    def __or__(self, other: Mapping[str, Sequence[str] | type[RepoG]]) -> Repos:
        return type(self)({**self, **other})

    def new(self, other: Mapping[str, Sequence[str] | type[RepoG]]) -> Repos:
        return self | other

    def items(self) -> ItemsView[str, type[RepoG]]:
        return self.__data.items()

    def keys(self) -> KeysView[str]:
        return self.__data.keys()

    def values(self) -> ValuesView[type[RepoG]]:
        return self.__data.values()

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.__data!r})"


DefaultRepoGs = Repos(
    dict(
        file=FileRepoG,
        copr=CoprRepoG,
        repo=SimpleRepoG,
        baseurl=BaseurlRepoG,
        mirrorlist=MirrorlistRepoG,
    )
)



# SPDX-FileCopyrightText: 2023 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import importlib.util
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from types import ModuleType

BACKEND = "libdnf5"
NEEDS = (
    "libdnf5",
    # Remove when https://github.com/rpm-software-management/dnf5/issues/281 is
    # resolved
    "rpm",
)


def ensure_backend() -> None:
    from fedrq.backends import MissingBackendError

    needs = ", ".join(NEEDS)
    for req in NEEDS:
        if not importlib.util.find_spec(req):
            raise MissingBackendError(f"Backend {BACKEND!r} requires {needs}")


def get_backend() -> ModuleType:
    from . import backend

    return backend



# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Provides access to the default package backend interface.
It is recommended to use the backend directly instead of this module.
"""

from __future__ import annotations

from collections.abc import Callable

from fedrq.backends import get_default_backend
from fedrq.backends.base import (
    BackendMod,
    BaseMakerBase,
    PackageCompat,
    PackageQueryCompat,
    RepoqueryBase,
)

backend: BackendMod = get_default_backend()
BaseMaker: type[BaseMakerBase] = backend.BaseMaker
Package: type[PackageCompat] = backend.Package
PackageQuery: type[PackageQueryCompat] = backend.PackageQuery
RepoError: Exception = backend.RepoError
Repoquery: type[RepoqueryBase] = backend.Repoquery
get_releasever: Callable[[], str] = backend.get_releasever
BACKEND: str = backend.BACKEND

__all__ = (
    "BACKEND",
    "BaseMaker",
    "Repoquery",
    "backend",
    "get_releasever",
)



# SPDX-FileCopyrightText: 2023 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later

"""
This module contains a fedrq backend (i.e. an implementation of the
[`fedrq.backends.base.BackendMod`][fedrq.backends.base.BackendMod] interface)
that uses the dnf Python bindings.
"""

from __future__ import annotations

import logging
import sys
import typing as t
from collections.abc import Collection
from functools import cache

from fedrq._utils import filter_latest
from fedrq.backends import MissingBackendError
from fedrq.backends.base import (
    BackendMod,
    BaseMakerBase,
    PackageCompat,
    PackageQueryCompat,
    RepoqueryBase,
)
from fedrq.backends.dnf import BACKEND

try:
    import dnf
    import hawkey
except ImportError:
    raise MissingBackendError from None

if t.TYPE_CHECKING:
    from _typeshed import StrPath

LOG = logging.getLogger(__name__)


class BaseMaker(BaseMakerBase):
    """
    Create a Base object and load repos
    """

    base: dnf.Base

    def __init__(self, base: dnf.Base | None = None) -> None:
        """
        Initialize and configure the base object.
        """
        self.base: dnf.Base = base or dnf.Base()

    @property
    def conf(self) -> dnf.conf.MainConf:
        return self.base.conf

    def set(self, key: str, value: t.Any) -> None:
        setattr(self.conf, key, value)

    def set_var(self, key: str, value: t.Any) -> None:
        if key not in self.base.conf.substitutions:
            raise KeyError(f"{key} is not a valid substituion")
        self.set(key, value)

    def fill_sack(
        self,
        *,
        from_cache: bool = False,
        load_system_repo: bool = False,
    ) -> dnf.Base:
        """
        Fill the sack and returns the dnf.Base object.
        The repository configuration shouldn't be manipulated after this.

        Note that the `_cachedir` arg is private and subject to removal.
        """
        if from_cache:
            self.base.fill_sack_from_repos_in_cache(load_system_repo=load_system_repo)
        else:
            self.base.fill_sack(load_system_repo=load_system_repo)
        return self.base

    def read_system_repos(self, disable: bool = True) -> None:
        """
        Load system repositories into the base object.
        By default, they are all disabled even if 'enabled=1' is in the
        repository configuration.
        """
        self.base.read_all_repos()
        if not disable:
            return None
        for repo in self.base.repos.iter_enabled():
            repo.disable()

    def enable_repos(self, repos: Collection[str]) -> None:
        """
        Enable a list of repositories by their repoid.
        Raise a ValueError if the repoid is not in `self.base`'s configuration.
        """
        for repo in repos:
            self.enable_repo(repo)

    def enable_repo(self, repo: str) -> None:
        """
        Enable a repo by its id.
        Raise a ValueError if the repoid is not in `self.base`'s configuration.
        """
        if repo_obj := self.base.repos.get_matching(repo):
            repo_obj.enable()
        else:
            raise ValueError(f"{repo} repo definition was not found.")

    def disable_repo(self, repo: str, ignore_missing: bool = True) -> None:
        """
        Disable a repo by its id.
        Raise a ValueError if the repoid is not in `self.base`'s configuration
        when ignore_missing is False.
        """
        if repo_obj := self.base.repos.get_matching(repo):
            repo_obj.disable()
        elif not ignore_missing:
            raise ValueError(f"{repo} repo definition was not found.")

    def read_repofile(self, file: StrPath) -> None:
        rr = dnf.conf.read.RepoReader(self.base.conf, None)
        for repo in rr._get_repos(str(file)):
            LOG.debug("Adding %s from %s", repo.id, file)
            self.base.repos.add(repo)

    # This is private for now
    def _read_repofile_new(self, file: StrPath, ensure_enabled: bool = False) -> None:
        """
        Load repositories from a repo file if they're not already in the
        configuration.
        """
        rr = dnf.conf.read.RepoReader(self.base.conf, None)
        for repo in rr._get_repos(str(file)):
            if repo.id in self.base.repos:
                LOG.debug("Not adding %s. It's already in the config.", repo.id)
            else:
                LOG.debug("Adding %s from %s", repo.id, file)
                self.base.repos.add(repo)
            if ensure_enabled:
                LOG.debug("Ensuring that %s is enabled.", repo.id)
                self.base.repos[repo.id].enable()

    def create_repo(self, repoid: str, **kwargs) -> None:
        """
        Add a Repo object to the repo sack and configure it.
        :param kwargs: key-values options that should be set on the Repo object
                       values (like $basearch) will be substituted automatically.
        """
        self.base.repos.add_new_repo(repoid, self.conf, **kwargs)

    @property
    def backend(self) -> BackendMod:
        return sys.modules[__name__]

    def repolist(self, enabled: bool | None = None) -> list[str]:
        if enabled is None:
            return list(self.base.repos)
        return [r.id for r in self.base.repos.values() if r.enabled is bool(enabled)]


class Repoquery(RepoqueryBase):
    def __init__(self, base: dnf.Base) -> None:
        self.base: dnf.Base = base

    @property
    def base_arches(self) -> set[str]:
        return {self.base.conf.arch, self.base.conf.basearch}

    def _query(self) -> dnf.query.Query:
        return self.base.sack.query()

    def resolve_pkg_specs(
        self,
        specs: Collection[str],
        resolve: bool = False,
        latest: int | None = None,
        with_src: bool = True,
    ) -> dnf.query.Query:
        ...
        LOG.debug(f"specs={specs}, resolve={resolve}, latest={latest}")
        query = self.query(empty=True)
        for p in specs:
            subject = dnf.subject.Subject(p).get_best_query(
                self.base.sack,
                with_provides=resolve,
                with_filenames=resolve,
                with_src=with_src,
            )
            query = query.union(subject)
            # LOG.debug(f"subject query: {tuple(subject)}")
        if resolve:
            query = query.union(self.query(provides=specs))
        filter_latest(query, latest)
        return query

    @property
    def backend(self) -> BackendMod:
        return sys.modules[__name__]


@cache
def get_releasever():
    """
    Return the system releasever
    """
    return dnf.rpm.detect_releasever("/")


Package: PackageCompat = dnf.package.Package
PackageQuery: PackageQueryCompat = dnf.query.Query
RepoError = dnf.exceptions.RepoError

__all__ = (
    "BACKEND",
    "BaseMaker",
    "Package",
    "PackageQuery",
    "RepoError",
    "Repoquery",
    "get_releasever",
    #
    "dnf",
    "hawkey",
)



# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import sys

from fedrq.cli import main

main(prog=f"{sys.executable} -m fedrq")



# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import logging
import typing as t
from collections.abc import MutableMapping

if t.TYPE_CHECKING:
    from fedrq.backends.base import PackageCompat, PackageQueryCompat

logger = logging.getLogger(__name__)
# PkgIter = t.Union[hawkey.Query, t.Iterable[dnf.package.Package]]


def get_source_name(package: PackageCompat) -> str:
    return package.name if package.arch == "src" else t.cast(str, package.source_name)


def filter_latest(query: PackageQueryCompat, latest: t.Optional[int]) -> None:
    # logger.debug("filter_latest(query={}, latest={})".format(tuple(query), latest))
    if latest:
        query.filterm(latest=latest)


def mklog(*args: str) -> logging.Logger:
    return logging.getLogger(".".join(args))


def merge_dict(
    data: MutableMapping[str, t.Any], dest: MutableMapping[str, t.Any]
) -> None:
    for key, value in tuple(data.items()):
        if key not in dest:
            pass
        elif isinstance(value, MutableMapping) and isinstance(
            dest[key], MutableMapping
        ):
            merge_dict(value, dest[key])
            continue
        dest[key] = value



# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later
"""
Generic tests for fedrq.cli.Command
"""
import os
import shutil
from pathlib import Path

import pytest
import tomli_w

import fedrq.cli
import fedrq.config

SUBCOMMANDS = ("pkgs", "whatrequires", "subpkgs")


# @pytest.mark.parametrize("subcommand", SUBCOMMANDS)
# def test_no_dnf_clean_failure(subcommand, capsys, monkeypatch):
#     monkeypatch.setattr(fedrq.cli.base, "HAS_DNF", False)
#     monkeypatch.setattr(fedrq.cli.base, "dnf", None)
#     monkeypatch.setattr(fedrq.cli.base, "hawkey", None)

#     with pytest.raises(SystemExit, match=r"^1$") as exc:
#         fedrq.cli.main([subcommand, "dummy"])
#     assert exc.value.code == 1
#     stdout, stderr = capsys.readouterr()
#     assert not stdout
#     assert stderr == "FATAL ERROR: " + fedrq.cli.base.NO_DNF_ERROR + "\n"


@pytest.mark.parametrize("subcommand", SUBCOMMANDS)
def test_smartcache_used(
    subcommand, mocker, monkeypatch, patch_config_dirs, temp_smartcache: Path
):
    """
    Ensure that the smartcache is used when the requested
    branch's releasever is different the the system's releasever
    """
    assert not list(temp_smartcache.iterdir())

    mocks = {}

    def _set_config(self, key: str) -> None:
        arg = getattr(self.args, key, None)
        if arg is not None:
            setattr(self.config, key, arg)
        if key == "backend":
            mocks["get_releasever"] = mocker.patch.object(
                self.config.backend_mod, "get_releasever", return_value="rawhide"
            )
            mocks["bm_set"] = mocker.spy(self.config.backend_mod.BaseMaker, "set")

    cls = fedrq.cli.COMMANDS[subcommand]
    monkeypatch.setattr(cls, "_set_config", _set_config)

    parser = cls.make_parser()
    args = parser.parse_args(["--sc", "packageb"])
    obj = cls(args)

    mocks["get_releasever"].assert_called_once()

    expected_cachedir = temp_smartcache / "fedrq" / "tester"

    assert any(
        call.args[1:] == ("cachedir", str(expected_cachedir))
        for call in mocks["bm_set"].call_args_list
    )

    assert obj.args.smartcache

    assert list(temp_smartcache.iterdir())


@pytest.mark.parametrize("subcommand", SUBCOMMANDS)
def test_smartcache_not_used(
    subcommand, mocker, monkeypatch, patch_config_dirs, temp_smartcache
):
    """
    Ensure that the smartcache is not used when the requested branch's
    releasever matches the the system's releasever
    """
    assert not list(temp_smartcache.iterdir())

    mocks = {}

    def _set_config(self, key: str) -> None:
        arg = getattr(self.args, key, None)
        if arg is not None:
            setattr(self.config, key, arg)
        if key == "backend":
            mocks["get_releasever"] = mocker.patch.object(
                self.config.backend_mod, "get_releasever", return_value="tester"
            )
            mocks["bm_set"] = mocker.spy(self.config.backend_mod.BaseMaker, "set")

    cls = fedrq.cli.COMMANDS[subcommand]
    monkeypatch.setattr(cls, "_set_config", _set_config)
    parser = cls.make_parser()
    args = parser.parse_args(["--sc", "packageb"])
    cls(args)

    mocks["get_releasever"].assert_called_once()
    assert all(call.args[1] != "cachedir" for call in mocks["bm_set"].call_args_list)
    # assert not list(temp_smartcache.iterdir())


@pytest.mark.parametrize(
    "args, config_smartcache, cachedir",
    (
        # smartcache is specified in the config file (default)
        ([], True, lambda d: d / "tester"),
        # smartcache is specified in the config file and on the cli (redundant)
        (["--sc"], True, lambda d: d / "tester"),
        # smartcache is only specified on the cli
        (["--sc"], False, lambda d: d / "tester"),
        # --system-cache is used to override the config file's 'smartcache = true'
        (["--system-cache"], False, lambda _: None),
        # --system-cache is used when smartcache is disabled in the config
        ([], False, lambda _: None),
        # --system-cache is used when smartcache is disabled in the config (redundant)
        (["--system-cache"], False, lambda _: None),
        # --cachedir trumps smartcache
        (["--cachedir=blah"], True, lambda _: Path("blah")),
        (["--cachedir=blah"], False, lambda _: Path("blah")),
    ),
)
def test_smartcache_config(
    args,
    config_smartcache,
    cachedir,
    patch_config_dirs,
    temp_smartcache,
    mocker,
    default_backend,
):
    assert os.environ["XDG_CACHE_HOME"] == str(temp_smartcache)
    bm_set = mocker.spy(default_backend.BaseMaker, "set")

    write_config = [True]
    # Check that True is the default
    if config_smartcache:
        write_config.append(False)
    dest = patch_config_dirs / "smartcache.toml"
    assert not dest.exists()
    for w in write_config:
        try:
            if w:
                data = {"smartcache": config_smartcache}
                with dest.open("wb") as fp:
                    tomli_w.dump(data, fp)
            parser = fedrq.cli.Pkgs.make_parser()
            pargs = parser.parse_args([*args, "packagea"])
            fedrq.cli.Pkgs(pargs)
            expected_cachedir = cachedir(temp_smartcache / "fedrq")
            if expected_cachedir:
                assert any(
                    call.args[1:] == ("cachedir", str(expected_cachedir))
                    for call in bm_set.call_args_list
                )
            else:
                assert all(call.args[1] != "cachedir" for call in bm_set.call_args_list)

        finally:
            shutil.rmtree("blah", ignore_errors=True)
            dest.unlink(True)


@pytest.mark.parametrize("subcommand", SUBCOMMANDS)
def test_nonexistant_formatter(subcommand, patch_config_dirs, capsys):
    with pytest.raises(SystemExit, match=r"^1$"):
        fedrq.cli.main([subcommand, "--formatter=blahblah", "*"])
    stdout, stderr = capsys.readouterr()
    assert not stdout
    assert stderr.splitlines() == [
        "ERROR: 'blahblah' is not a valid formatter",
        fedrq.cli.base.FORMATTER_ERROR_SUFFIX,
    ]


@pytest.mark.parametrize("subcommand", SUBCOMMANDS)
@pytest.mark.parametrize("formatter", (("json"), ("attr")))
def test_formatter_0_args(subcommand, formatter, patch_config_dirs, capsys):
    with pytest.raises(SystemExit, match=r"^1$"):
        fedrq.cli.main([subcommand, "--formatter", formatter + ":", "*"])
    stdout, stderr = capsys.readouterr()
    assert not stdout
    assert stderr.splitlines() == [
        f"ERROR: '{formatter}' FormatterError: received less than 1 argument",
        fedrq.cli.base.FORMATTER_ERROR_SUFFIX,
    ]


@pytest.mark.parametrize("subcommand", SUBCOMMANDS)
def test_json_formatter_invalid_args(subcommand, patch_config_dirs, capsys):
    with pytest.raises(SystemExit, match=r"^1$"):
        fedrq.cli.main([subcommand, "-F", "json:abc,name,requires,xyz", "*"])
    stdout, stderr = capsys.readouterr()
    assert not stdout
    assert stderr.splitlines() == [
        "ERROR: 'json' FormatterError: invalid argument 'abc'",
        fedrq.cli.base.FORMATTER_ERROR_SUFFIX,
    ]



# SPDX-FileCopyrightText: 2023 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import abc
import importlib.resources
import logging
from collections.abc import Callable, Collection, Iterable, Iterator
from typing import TYPE_CHECKING, Any, Optional, Protocol, TypeVar, runtime_checkable
from warnings import warn

if TYPE_CHECKING:
    from _typeshed import StrPath

    from fedrq.config import Release

_QueryT = TypeVar("_QueryT", bound="PackageQueryCompat")
LOG = logging.getLogger("fedrq.backends")


@runtime_checkable
class PackageCompat(Protocol):
    """
    Common interface provided by dnf.package.Package and other backends
    """

    @property
    def name(self) -> str:
        ...

    @property
    def arch(self) -> str:
        ...

    @property
    def a(self) -> str:
        ...

    @property
    def epoch(self) -> int:
        ...

    @property
    def e(self) -> int:
        ...

    @property
    def version(self) -> str:
        ...

    @property
    def v(self) -> str:
        ...

    @property
    def release(self) -> str:
        ...

    @property
    def r(self) -> str:
        ...

    @property
    def from_repo(self) -> str:
        ...

    @property
    def evr(self) -> str:
        ...

    @property
    def debug_name(self) -> str:
        ...

    @property
    def source_name(self) -> Optional[str]:
        ...

    @property
    def source_debug_name(self) -> str:
        ...

    @property
    def installtime(self) -> int:
        ...

    @property
    def buildtime(self) -> int:
        ...

    @property
    def size(self) -> int:
        ...

    @property
    def downloadsize(self) -> int:
        ...

    @property
    def installsize(self) -> int:
        ...

    @property
    def provides(self) -> Iterable:
        ...

    @property
    def requires(self) -> Iterable:
        ...

    @property
    def recommends(self) -> Iterable:
        ...

    @property
    def suggests(self) -> Iterable:
        ...

    @property
    def supplements(self) -> Iterable:
        ...

    @property
    def enhances(self) -> Iterable:
        ...

    @property
    def obsoletes(self) -> Iterable:
        ...

    @property
    def conflicts(self) -> Iterable:
        ...

    @property
    def sourcerpm(self) -> Optional[str]:
        ...

    @property
    def description(self) -> str:
        ...

    @property
    def summary(self) -> str:
        ...

    @property
    def license(self) -> str:
        ...

    @property
    def url(self) -> str:
        ...

    @property
    def reason(self) -> Optional[str]:
        ...

    @property
    def files(self) -> Iterable[str]:
        ...

    @property
    def reponame(self) -> str:
        ...

    @property
    def repoid(self) -> str:
        ...

    @property
    def vendor(self) -> str:
        ...

    @property
    def packager(self) -> str:
        ...

    @property
    def location(self) -> str:
        ...

    def __hash__(self) -> int:
        ...

    def __lt__(self, other) -> bool:
        ...

    def __le__(self, other) -> bool:
        ...

    def __gt__(self, other) -> bool:
        ...

    def __ge__(self, other) -> bool:
        ...


@runtime_checkable
class PackageQueryCompat(Protocol):
    """
    Common PackageQuery interface provided by hawkey.Query and other backends.
    """

    def filter(self, **kwargs) -> PackageQueryCompat:
        """
        Filter the PackageQuery.
        Depending on the backend, this either modifies 'self' in place and
        return 'self' or return a new PackageQuery object.
        See https://dnf.readthedocs.io/en/latest/api_queries.html#dnf.query.Query.filter
        for the allowed kwargs.
        """
        ...

    def filterm(self, **kwargs) -> PackageQueryCompat:
        """
        Filter the PackageQuery in place and return 'self'.
        See https://dnf.readthedocs.io/en/latest/api_queries.html#dnf.query.Query.filter
        for the allowed kwargs.
        """
        ...

    def union(self: _QueryT, other: _QueryT) -> _QueryT:
        """
        Combine two PackageQuery objects.
        Depending on the backend, this either modifies 'self' in place and
        returns 'self' or returns a new PackageQuery object.
        """
        ...

    def __len__(self) -> int:
        ...

    def __iter__(self) -> Iterator[PackageCompat]:
        ...


class BaseMakerBase(abc.ABC):
    """
    Create a Base object, set configuration, and load repos
    """

    base: Any

    def __init__(self, base=None) -> None:
        self.base = base

    @abc.abstractmethod
    def fill_sack(
        self,
        *,
        from_cache: bool = False,
        load_system_repo: bool = False,
    ):
        """
        Fill the sack and returns the Base object.
        The repository configuration shouldn't be manipulated after this.
        'from_cache' isn't currently supported by the libdnf5 backend.
        """
        ...

    @abc.abstractmethod
    def read_system_repos(self, disable: bool = True) -> None:
        """
        Load system repositories into the base object.
        By default, they are all disabled even if 'enabled=1' is in the
        repository configuration.
        """

    @abc.abstractmethod
    def enable_repos(self, repos: Collection[str]) -> None:
        """
        Enable a list of repositories by their repoid.
        Raise a ValueError if the repoid is not in `self.base`'s configuration.
        """

    @abc.abstractmethod
    def enable_repo(self, repo: str) -> None:
        """
        Enable a repo by its id.
        Raise a ValueError if the repoid is not in `self.base`'s configuration.
        """

    @abc.abstractmethod
    def disable_repo(self, repo: str, ignore_missing: bool = True) -> None:
        """
        Disable a repo by its id.
        Raise a ValueError if the repoid is not in `self.base`'s configuration
        when ignore_missing is False.
        """

    @abc.abstractmethod
    def read_repofile(self, file: StrPath) -> None:
        """
        Load repositories from a repo file
        """

    @abc.abstractmethod
    def set(self, key: str, value: Any) -> None:
        """
        Set configuration options. Must be called before reading repos.
        """
        ...

    @abc.abstractmethod
    def set_var(self, key: str, value: Any) -> None:
        """
        Set substitutions (e.g. arch, basearch, releasever).
        Needs to be called before reading repos.
        """
        ...

    # Private for now
    @abc.abstractmethod
    def _read_repofile_new(self, file: StrPath, ensure_enabled: bool = False) -> None:
        """
        Load repositories from a repo file if they're not already in the
        configuration.
        """

    def sets(self, conf: dict[str, Any], substitutions: dict[str, Any]) -> None:
        """
        :param conf: A dict of configuration options. Call self.set() for each
        k-v pair.
        :param substitutions: A dict of substitutions/vars options. Call self.set_var()
        for each k-v pair.
        """
        for opt in conf.items():
            self.set(*opt)
        for opt in substitutions.items():
            self.set_var(*opt)

    def load_filelists(self) -> None:
        # Can be overriden by subclasses. Purposely isn't an @abstractmethod.
        """
        Load the filelists if they're not already enabled default
        """
        return None

    def load_release_repos(self, release: Release, set_releasever: bool = True) -> None:
        """
        Load the repositories from a fedrq.config.Release object
        """
        if set_releasever:
            self.set_var("releasever", release.version)
        if release.release_config.system_repos:
            self.read_system_repos(
                disable=not release.release_config.append_system_repos
            )
        for path in release.release_config.full_def_paths:
            with importlib.resources.as_file(path) as fp:
                LOG.debug("Reading %s", fp)
                self._read_repofile_new(fp)
        release.repog.load(self, release.config, release)

    @abc.abstractmethod
    def create_repo(self, repoid: str, **kwargs) -> None:
        """
        Add a Repo object to the repo sack and configure it.
        :param kwargs: key-values options that should be set on the Repo object
                       values (like $basearch) will be substituted automatically.
        """
        ...

    @property
    @abc.abstractmethod
    def backend(self) -> BackendMod:
        ...

    @abc.abstractmethod
    def repolist(self, enabled: bool | None = None) -> list[str]:
        ...


class RepoqueryBase(abc.ABC):
    """
    Helpers to query a repository.
    Provides a unified repoquery interface for different backends.
    """

    def __init__(self, base) -> None:
        self.base = base

    @property
    @abc.abstractmethod
    def base_arches(self) -> set[str]:
        """
        Return a set of the system's arch and basearch.
        """
        ...

    @abc.abstractmethod
    def resolve_pkg_specs(
        self,
        specs: Collection[str],
        resolve: bool = False,
        latest: int | None = None,
        with_src: bool = True,
    ) -> PackageQueryCompat:
        """
        Resolve pkg specs.
        See
        https://dnf.readthedocs.io/en/latest/command_ref.html?highlight=spec#specifying-packages
        or
        https://dnf5.readthedocs.io/en/latest/misc/specs.7.html
        for valid forms.

        :param specs: Package specs to resolve.
        :param resolve: Whether to resolve file paths or virtual Provides in
                        addition to package specs
        :param latest: Limit packages with the same name and arch.
        """
        ...

    def arch_filterm(
        self, query: PackageQueryCompat, arch: str | Iterable[str] | None = None
    ) -> PackageQueryCompat:
        """
        Filter a query's architectures in place and return it.
        It includes a little more functionality than query.filterm(arch=...).

        - When arch is None, the query is left untouched.
        - If arch equals 'notsrc', all src and multilib packages are
          excluded.
        - If arch equals 'arched', all noarch, multilib, and source
          packages are excluded.
        - Otherwise, arch is passed to query.filterm(arch=...) and no other
          validation is preformed.
        """
        if not arch:
            return query
        if arch == "notsrc":
            return query.filterm(arch=(*self.base_arches, "noarch"))  # type: ignore
        elif arch == "arched":
            return query.filterm(arch=self.base.conf.basearch)
        else:
            return query.filterm(arch=arch)

    def arch_filter(
        self, query: PackageQueryCompat, arch: str | Iterable[str] | None = None
    ) -> PackageQueryCompat:
        """
        Filter a query's architectures and return it.
        It includes a little more functionality than query.filter(arch=...).

        - When arch is None, the query is left untouched.
        - If arch equals 'notsrc', all src and multilib packages are
          excluded.
        - If arch equals 'arched', all noarch, multilib, and source
          packages are excluded.
        - Otherwise, arch is passed to query.filterm(arch=...) and no other
          validation is preformed.
        """
        if not arch:
            return query
        if arch == "notsrc":
            return query.filter(arch=(*self.base_arches, "noarch"))  # type: ignore
        if arch == "arched":
            return query.filter(arch=list(self.base_arches))
        return query.filter(arch=arch)

    @abc.abstractmethod
    def _query(self) -> PackageQueryCompat:
        """
        Return the PackageQuery object for this backend
        """
        return self.base.sack.query()

    def query(
        self, *, arch: str | Iterable[str] | None = None, **kwargs
    ) -> PackageQueryCompat:
        """
        Return an inital PackageQuery that's filtered with **kwargs.
        Further filtering can be applied with the PackageQuery's filter and
        filterm methods.
        """
        if kwargs.get("latest") is None:
            kwargs.pop("latest", None)
        query = self._query()
        query.filterm(**kwargs)
        self.arch_filterm(query, arch)
        return query

    def get_package(
        self,
        name: str,
        arch: str | Iterable[str] | None = None,
    ) -> PackageCompat:
        """
        Return the latest Package that matches the 'name' and 'arch'.
        A ValueError is raised when no matches are found.
        """
        query = self.query(name=name, arch=arch, latest=1)
        if len(query) < 1:
            raise ValueError(f"Zero packages found for {name} on {arch}")
        return next(iter(query))

    def get_subpackages(
        self, packages: Iterable[PackageCompat], **kwargs
    ) -> PackageQueryCompat:
        """
        Return a PackageQuery containing the binary RPMS/subpackages produced
        by {packages}.

        :param package: A :class:`PackageQueryCompat` containing source packages
        """
        arch = kwargs.get("arch")
        if arch == "src":
            raise ValueError("{arch} cannot be 'src'")
        elif not arch:
            kwargs.setdefault("arch__neq", "src")
        if val := kwargs.pop("sourcerpm", None):
            warn(f"Removing invalid kwarg: 'sourcerpm={val}")

        for package in packages:
            if package.arch != "src":
                raise ValueError(f"{package} must be a source package.")

        sourcerpms = [
            f"{package.name}-{package.version}-{package.release}.src.rpm"
            for package in packages
        ]
        query = self.query(sourcerpm=sourcerpms, **kwargs)
        return query

    @property
    @abc.abstractmethod
    def backend(self) -> BackendMod:
        ...


class BackendMod(Protocol):
    """
    Protocol for a fedrq backend module.
    Each backend module (e.g.
    [`fedrq.backends.dnf.backend`][fedrq.backends.dnf.backend])
    implements this interface.
    """

    BACKEND: str
    BaseMaker: type[BaseMakerBase]
    Package: type[PackageCompat]
    PackageQuery: type[PackageQueryCompat]
    Repoquery: type[RepoqueryBase]
    RepoError: Exception
    get_releasever: Callable[[], str]
